# Copyright Least Authority Enterprises.
# See LICENSE for details.

"""
Kubernetes authentication support.
"""

import pem

from OpenSSL.crypto import FILETYPE_PEM, Error as OpenSSLError

from zope.interface import implementer

from pyrsistent import PClass, field, pmap_field

from twisted.python.url import URL
from twisted.internet import ssl
from twisted.web.iweb import IPolicyForHTTPS, IAgent
from twisted.web.http_headers import Headers
from twisted.web.client import Agent

from pykube import KubeConfig

from ._invariants import instance_of

class TLSCredentials(PClass):
    """
    ``TLSCredentials`` holds the information necessary to use a client
    certificate for a TLS handshake.

    :ivar pem.Certificate certificate: The client certificate to use.
    :ivar pem.Key key: The private key which corresponds to ``certificate``.
    """
    certificate = field(mandatory=True, invariant=instance_of(pem.Certificate))
    key = field(mandatory=True, invariant=instance_of(pem.Key))


class NetLocation(PClass):
    """
    ``NetLocation`` holds information which identifies a particular HTTPS
    server.  This is useful as a key for selecting the right certificate
    authority and client certificate to use.

    :ivar unicode host: The server's hostname.
    :ivar port: The server's port number.
    """
    host = field(mandatory=True, type=unicode)
    port = field(mandatory=True, type=(int, long))


def pick_cert_for_twisted(netloc, possible):
    """
    Pick the right client key/certificate to use for the given server and
    return it in the form Twisted wants.

    :param NetLocation netloc: The location of the server to consider.
    :param dict[TLSCredentials] possible: The available credentials from which
        to choose.

    :return: A ``twisted.internet.ssl.PrivateCertificate`` instance
        representing them if credentials are found.  Otherwise, ``None``.
    """
    try:
        creds = possible[netloc]
    except KeyError:
        return None

    key = ssl.KeyPair.load(creds.key.as_bytes(), FILETYPE_PEM)
    return ssl.PrivateCertificate.load(creds.certificate.as_bytes(), key, FILETYPE_PEM)


def pick_trust_for_twisted(netloc, possible):
    """
    Pick the right "trust roots" (certificate authority certificates) for the
    given server and return it in the form Twisted wants.

    Kubernetes certificates are often self-signed or otherwise exist outside
    of the typical certificate authority cartel system common for normal
    websites.  This function tries to find the right authority to use.

    :param NetLocation netloc: The location of the server to consider.
    :param dict[pem.Certificate] possible: The available certificate authority
        certificates from which to choose.

    :return: A provider of ``twisted.internet.interfaces.IOpenSSLTrustRoot``
    if there is a known certificate authority certificate for the given
    server.  Otherwise, ``None``.
    """
    try:
        trust_cert = possible[netloc]
    except KeyError:
        return None

    cert = ssl.Certificate.load(trust_cert.as_bytes(), FILETYPE_PEM)
    return ssl.trustRootFromCertificates([cert])


@implementer(IPolicyForHTTPS)
class ClientCertificatePolicyForHTTPS(PClass):
    """
    ``ClientCertificatePolicyForHTTPS`` selects the correct client certificate
    and trust roots to use for interacting with the Kubernetes API server.

    :ivar credentials: All available client certificates.
    :ivar trust_roots: All available certificate authority certificates.
    """
    credentials = pmap_field(
        NetLocation, TLSCredentials,
    )

    trust_roots = pmap_field(
        NetLocation, pem.Certificate,
    )

    def creatorForNetloc(self, hostname, port):
        """
        Pick from amongst client certs and ca certs to create a proper TLS context
        factory.

        :see: ``twisted.web.iweb.IPolicyForHTTPS``
        """
        hostname = hostname.decode("ascii")

        netloc = NetLocation(host=hostname, port=port)
        client_cert = pick_cert_for_twisted(netloc, self.credentials)
        trust_root = pick_trust_for_twisted(netloc, self.trust_roots)

        return ssl.optionsForClientTLS(
            # It is not necessarily the case that the certificate presented
            # will use this name but it is common to encounter self-signed
            # certificates which do use this name.  There doesn't seem to be
            # anything in the configuration file which would tell us what the
            # proper name is.  We'll probably need to make this configurable
            # at some point, I guess.
            u"kubernetes",
            clientCertificate=client_cert,
            trustRoot=trust_root,
        )


def authenticate_with_certificate(reactor, base_url, client_cert, client_key, ca_cert):
    """
    Create an ``IAgent`` which can issue authenticated requests to a
    particular Kubernetes server using a client certificate.

    :param reactor: The reactor with which to configure the resulting agent.

    :param twisted.python.url.URL base_url: The base location of the
        Kubernetes API.

    :param pem.Certificate client_cert: The client certificate to use.

    :param pem.Key client_key: The private key to use with the client
        certificate.

    :param pem.Certificate ca_cert: The certificate authority to respect when
        verifying the Kubernetes server certificate.

    :return IAgent: An agent which will authenticate itself to a particular
        Kubernetes server and which will verify that server or refuse to
        interact with it.
    """
    if base_url.scheme != u"https":
        raise ValueError(
            "authenticate_with_certificate() makes sense for HTTPS, not {!r}".format(
                base_url.scheme
            ),
        )

    netloc = NetLocation(host=base_url.host, port=base_url.port)
    policy = ClientCertificatePolicyForHTTPS(
        credentials={
            netloc: TLSCredentials(certificate=client_cert, key=client_key),
        },
        trust_roots={
            netloc: ca_cert,
        },
    )
    return Agent(reactor, contextFactory=policy)



@implementer(IAgent)
class HeaderInjectingAgent(PClass):
    """
    An ``IAgent`` which adds some headers to every request it makes.

    :ivar Headers _to_inject: The headers to add.
    @ivar IAgent _agent: The agent to use to issue requests.
    """
    _to_inject = field(mandatory=True)
    _agent = field(mandatory=True)

    def request(self, method, url, headers=None, bodyProducer=None):
        """
        Issue a request with some extra headers.

        :see: ``twisted.web.iweb.IAgent.request``
        """
        if headers is None:
            headers = Headers()
        else:
            headers = headers.copy()
        for k, vs in self._to_inject.getAllRawHeaders():
            headers.setRawHeaders(k, vs)
        return self._agent.request(method, url, headers, bodyProducer)



def authenticate_with_serviceaccount(reactor, **kw):
    """
    Create an ``IAgent`` which can issue authenticated requests to a
    particular Kubernetes server using a service account token.

    :param reactor: The reactor with which to configure the resulting agent.

    :param bytes path: The location of the service account directory.  The
        default should work fine for normal use within a container.

    :return IAgent: An agent which will authenticate itself to a particular
        Kubernetes server and which will verify that server or refuse to
        interact with it.
    """
    config = KubeConfig.from_service_account(**kw)

    token = config.user["token"]
    base_url = URL.fromText(config.cluster["server"].decode("ascii"))

    ca_certs = pem.parse(config.cluster["certificate-authority"].bytes())
    if not ca_certs:
        raise ValueError("No certificate authority certificate found.")
    ca_cert = ca_certs[0]

    try:
        # Validate the certificate so we have early failures for garbage data.
        ssl.Certificate.load(ca_cert.as_bytes(), FILETYPE_PEM)
    except OpenSSLError as e:
        raise ValueError(
            "Invalid certificate authority certificate found.",
            str(e),
        )

    netloc = NetLocation(host=base_url.host, port=base_url.port)
    policy = ClientCertificatePolicyForHTTPS(
        credentials={},
        trust_roots={
            netloc: ca_cert,
        },
    )

    agent = HeaderInjectingAgent(
        _to_inject=Headers({u"authorization": [u"Bearer {}".format(token)]}),
        _agent=Agent(reactor, contextFactory=policy),
    )
    return agent



# Copyright Least Authority Enterprises.
# See LICENSE for details.

"""
Functional structures for representing different kinds of Kubernetes
state.
"""

from uuid import uuid4

from zope.interface import implementer

from pyrsistent import mutant

from twisted.python.filepath import FilePath

from . import UnrecognizedVersion, UnrecognizedKind, IObject
from ._swagger import Swagger, VersionedPClasses

spec = Swagger.from_path(FilePath(__file__).sibling(u"kubernetes-1.5.json"))
v1 = VersionedPClasses(
    spec, u"v1", name_field=u"kind", version_field=u"apiVersion",
)
v1beta1 = VersionedPClasses(
    spec, u"v1beta1", name_field=u"kind", version_field=u"apiVersion",
)


def behavior(namespace):
    """
    Create a class decorator which adds the resulting class to the given
    namespace-y thing.
    """
    def decorator(cls):
        setattr(namespace, cls.__name__, cls)
        return cls
    return decorator



@behavior(v1)
class NamespaceStatus(v1.NamespaceStatus):
    """
    ``NamespaceStatus`` instances model `Kubernetes namespace status
    <https://kubernetes.io/docs/api-reference/v1/definitions/#_v1_namespacestatus>`_.
    """
    @classmethod
    def active(cls):
        return cls(phase=u"Active")


    @classmethod
    def terminating(cls):
        return cls(phase=u"Terminating")



@behavior(v1)
@implementer(IObject)
class Namespace(v1.Namespace):
    """
    ``Namespace`` instances model `Kubernetes namespaces
    <https://kubernetes.io/docs/user-guide/namespaces/>`_.
    """
    @classmethod
    def default(cls):
        """
        Get the default namespace.
        """
        return cls(metadata=v1.ObjectMeta(name=u"default"))


    def fill_defaults(self):
        return self.transform(
            # TODO Also creationTimestamp, resourceVersion, maybe selfLink.
            # Also, should this clobber existing values or leave them alone?
            # See https://github.com/LeastAuthority/txkube/issues/36
            [u"metadata", u"uid"], unicode(uuid4()),
            [u"status"], NamespaceStatus.active(),
        )


    def delete_from(self, collection):
        # TODO: deletionTimestamp?  Terminating status?  See #24
        return collection.remove(self)



@behavior(v1)
@implementer(IObject)
class ConfigMap(v1.ConfigMap):
    """
    ``ConfigMap`` instances model `ConfigMap objects
    <https://kubernetes.io/docs/api-reference/v1/definitions/#_v1_configmap>`_.
    """
    def fill_defaults(self):
        # TODO Surely some stuff to fill.
        # See https://github.com/LeastAuthority/txkube/issues/36
        return self


    def delete_from(self, collection):
        return collection.remove(self)



@behavior(v1)
@implementer(IObject)
class Service(v1.Service):
    """
    ``Service`` instances model `Service objects
    <https://kubernetes.io/docs/api-reference/v1/definitions/#_v1_service>`_.
    """
    def fill_defaults(self):
        # TODO Surely some stuff to fill.
        # See https://github.com/LeastAuthority/txkube/issues/36
        return self


    def delete_from(self, collection):
        return collection.remove(self)



def set_if_none(desired_value):
    """
    Create a transformer which sets the given value if it finds ``None`` as
    the current value, otherwise leaves the current value alone.
    """
    def transform(current_value):
        if current_value is None:
            return desired_value
        return current_value
    return transform



@behavior(v1beta1)
@implementer(IObject)
class Deployment(v1beta1.Deployment):
    """
    ``Deployment`` instances model `Deployment objects
    <https://kubernetes.io/docs/api-reference/extensions/v1beta1/definitions/#_v1beta1_deployment>`_.
    """
    def fill_defaults(self):
        # Copying apparent Kubernetes behavior.
        return self.transform(
            [u"metadata", u"labels"],
            set_if_none(self.spec.template.metadata.labels),
        )


    def delete_from(self, collection):
        return collection.remove(self)



@behavior(v1beta1)
@implementer(IObject)
class ReplicaSet(v1beta1.ReplicaSet):
    """
    ``ReplicaSet`` instances model `ReplicaSet objects
    <https://kubernetes.io/docs/concepts/workloads/controllers/replicaset/>`_.
    """
    def fill_defaults(self):
        return self


    def delete_from(self, collection):
        return collection.remove(self)



@behavior(v1)
@implementer(IObject)
class Pod(v1.Pod):
    """
    ``Pod`` instances model `Pod objects
    <https://kubernetes.io/docs/api-reference/v1/definitions/#_v1_pod>`_.
    """
    def fill_defaults(self):
        return self


    def delete_from(self, collection):
        return collection.remove(self)



def object_sort_key(obj):
    """
    Define a predictable sort ordering for Kubernetes objects.

    This should be the same ordering that Kubernetes itself imposes.
    """
    return (
        # Not all objects have a namespace.
        getattr(obj.metadata, "namespace", None),
        obj.metadata.name,
    )



def required_unique(objects, key):
    """
    A pyrsistent invariant which requires all objects in the given iterable to
    have a unique key.

    :param objects: The objects to check.
    :param key: A one-argument callable to compute the key of an object.

    :return: An invariant failure if any two or more objects have the same key
        computed.  An invariant success otherwise.
    """
    keys = {}
    duplicate = set()
    for k in map(key, objects):
        keys[k] = keys.get(k, 0) + 1
        if keys[k] > 1:
            duplicate.add(k)
    if duplicate:
        return (False, u"Duplicate object keys: {}".format(duplicate))
    return (True, u"")



def add(value):
    def evolver(pvector):
        return sorted(pvector.append(value), key=object_sort_key)
    return evolver


def remove(value):
    def evolver(pset):
        return pset.remove(value)
    return evolver



class _List(object):
    def __invariant__(self):
        return required_unique(self.items, object_sort_key)


    @staticmethod
    def __new__(cls, **kwargs):
        # The Kubernetes Swagger schema for Lists claims items is a *required*
        # *array*.  However, it is frequently None/null instead.  Hack around
        # such values, turning them into the empty sequence.
        #
        # It might be better to fix this with a schema overlay instead - eg,
        # with a tweak to mark them as optional instead of required.
        items = kwargs.get(u"items", None)
        if items is None:
            kwargs[u"items"] = ()
        return super(_List, cls).__new__(cls, **kwargs)


    def item_by_name(self, name):
        """
        Find an item in this collection by its name metadata.

        :param unicode name: The name of the object for which to search.

        :raise KeyError: If no object matching the given name is found.
        :return IObject: The object with the matching name.
        """
        for obj in self.items:
            if obj.metadata.name == name:
                return obj
        raise KeyError(name)


    def add(self, obj):
        return self.transform([u"items"], add(obj))


    def remove(self, obj):
        return self.transform([u"items"], remove(obj))


    def replace(self, old, new):
        return self.transform(
            [u"items"], remove(old),
            [u"items"], add(new),
        )



@behavior(v1)
@implementer(IObject)
class NamespaceList(_List, v1.NamespaceList):
    pass



@behavior(v1)
@implementer(IObject)
class ConfigMapList(_List, v1.ConfigMapList):
    pass



@behavior(v1)
@implementer(IObject)
class ServiceList(_List, v1.ServiceList):
    pass



@behavior(v1beta1)
@implementer(IObject)
class DeploymentList(_List, v1beta1.DeploymentList):
    pass



@behavior(v1beta1)
@implementer(IObject)
class ReplicaSetList(_List, v1beta1.ReplicaSetList):
    pass



@behavior(v1)
@implementer(IObject)
class PodList(_List, v1.PodList):
    pass



_versions = {
    u"v1": v1,
    u"v1beta1": v1beta1,
}

def _mutilate(version):
    if version == u"v1beta1":
        return u"extensions/v1beta1"
    return version



def _unmutilate(version):
    if version.startswith(u"extensions/"):
        return version[len(u"extensions/"):]
    return version



def iobject_to_raw(obj):
    result = obj.serialize()
    result.update({
        u"kind": obj.kind,
        u"apiVersion": _mutilate(obj.apiVersion),
    })
    return result

@mutant
def iobject_from_raw(obj):
    """
    Load an object of unspecified type from the raw representation of it.

    :raise KeyError: If the kind of object is unsupported.

    :return IObject: The loaded object.
    """
    kind = obj[u"kind"]
    apiVersion = _unmutilate(obj[u"apiVersion"])
    try:
        v = _versions[apiVersion]
    except KeyError:
        raise UnrecognizedVersion(apiVersion, obj)
    try:
        cls = getattr(v, kind)
    except AttributeError:
        raise UnrecognizedKind(apiVersion, kind, obj)
    others = obj.discard(u"kind").discard(u"apiVersion")
    return cls.create(others)



# Copyright Least Authority Enterprises.
# See LICENSE for details.

"""
Integration between Eliot, eliottree, and testtools to provide easily
readable Eliot logs for failing tests.
"""

from io import BytesIO

from fixtures import Fixture

from eliot import add_destination, remove_destination
from eliottree import Tree, render_task_nodes

from testtools.content import Content
from testtools.content_type import UTF8_TEXT


def _eliottree(logs):
    """
    Render some Eliot log events into a tree-like string.

    :param list[dict] logs: The Eliot log events to render.  These should be
        dicts like those passed to an Eliot destination.

    :return bytes: The rendered string.
    """
    tree = Tree()
    tree.merge_tasks(logs)
    nodes = tree.nodes()

    out = BytesIO()
    render_task_nodes(
        write=out.write,
        nodes=nodes,
        field_limit=0,
    )
    return out.getvalue()



class CaptureEliotLogs(Fixture):
    """
    A fixture which captures Eliot logs emitted while it is active and adds a
    detail which includes the (easily human-readable) "tree" rendering of
    those logs.
    """
    LOG_DETAIL_NAME = "eliot-log"

    # Unusually, the Fixtures convention is that underscore prefixed methods
    # are not private.  Instead, they're more like internal hooks.  It is
    # intended that application code override these methods you might
    # otherwise expect are private.
    def _setUp(self):
        self.logs = []
        add_destination(self.logs.append)
        self.addCleanup(lambda: remove_destination(self.logs.append))
        self.addDetail(
            self.LOG_DETAIL_NAME,
            Content(
                UTF8_TEXT,
                # Safeguard the logs against _tearDown.  Capture the list
                # object in the lambda's defaults.
                lambda logs=self.logs: [_eliottree(logs)],
            ),
        )



# Copyright Least Authority Enterprises.
# See LICENSE for details.

import os

from uuid import uuid4

from fixtures import TempDir

from testtools.matchers import Equals, Contains, raises

from twisted.python.filepath import FilePath
from twisted.internet.protocol import Factory
from twisted.web.http_headers import Headers
from twisted.test.iosim import ConnectionCompleter
from twisted.test.proto_helpers import AccumulatingProtocol, MemoryReactor

from ..testing import TestCase

from .. import authenticate_with_serviceaccount

# Just an arbitrary certificate pulled off the internet.  Details ought not
# matter.  Retrieved using:
#
#    $ openssl s_client -showcerts -connect google.com:443
#
_CA_CERT_PEM = b"""\
-----BEGIN CERTIFICATE-----
MIIDfTCCAuagAwIBAgIDErvmMA0GCSqGSIb3DQEBBQUAME4xCzAJBgNVBAYTAlVT
MRAwDgYDVQQKEwdFcXVpZmF4MS0wKwYDVQQLEyRFcXVpZmF4IFNlY3VyZSBDZXJ0
aWZpY2F0ZSBBdXRob3JpdHkwHhcNMDIwNTIxMDQwMDAwWhcNMTgwODIxMDQwMDAw
WjBCMQswCQYDVQQGEwJVUzEWMBQGA1UEChMNR2VvVHJ1c3QgSW5jLjEbMBkGA1UE
AxMSR2VvVHJ1c3QgR2xvYmFsIENBMIIBIjANBgkqhkiG9w0BAQEFAAOCAQ8AMIIB
CgKCAQEA2swYYzD99BcjGlZ+W988bDjkcbd4kdS8odhM+KhDtgPpTSEHCIjaWC9m
OSm9BXiLnTjoBbdqfnGk5sRgprDvgOSJKA+eJdbtg/OtppHHmMlCGDUUna2YRpIu
T8rxh0PBFpVXLVDviS2Aelet8u5fa9IAjbkU+BQVNdnARqN7csiRv8lVK83Qlz6c
JmTM386DGXHKTubU1XupGc1V3sjs0l44U+VcT4wt/lAjNvxm5suOpDkZALeVAjmR
Cw7+OC7RHQWa9k0+bw8HHa8sHo9gOeL6NlMTOdReJivbPagUvTLrGAMoUgRx5asz
PeE4uwc2hGKceeoWMPRfwCvocWvk+QIDAQABo4HwMIHtMB8GA1UdIwQYMBaAFEjm
aPkr0rKV10fYIyAQTzOYkJ/UMB0GA1UdDgQWBBTAephojYn7qwVkDBF9qn1luMrM
TjAPBgNVHRMBAf8EBTADAQH/MA4GA1UdDwEB/wQEAwIBBjA6BgNVHR8EMzAxMC+g
LaArhilodHRwOi8vY3JsLmdlb3RydXN0LmNvbS9jcmxzL3NlY3VyZWNhLmNybDBO
BgNVHSAERzBFMEMGBFUdIAAwOzA5BggrBgEFBQcCARYtaHR0cHM6Ly93d3cuZ2Vv
dHJ1c3QuY29tL3Jlc291cmNlcy9yZXBvc2l0b3J5MA0GCSqGSIb3DQEBBQUAA4GB
AHbhEm5OSxYShjAGsoEIz/AIx8dxfmbuwu3UOx//8PDITtZDOLC5MH0Y0FWDomrL
NhGc6Ehmo21/uBPUR/6LWlxz/K7ZGzIZOKuXNBSqltLroxwUCEm2u+WR74M26x1W
b8ravHNjkOR/ez4iyz0H7V84dJzjA1BOoa+Y7mHyhD8S
-----END CERTIFICATE-----
"""

class AuthenticateWithServiceAccountTests(TestCase):
    """
    Tests for ``authenticate_with_serviceaccount``.
    """
    def _authorized_request(self, token, headers):
        """
        Get an agent using ``authenticate_with_serviceaccount`` and issue a
        request with it.

        :return bytes: The bytes of the request the agent issues.
        """
        server = AccumulatingProtocol()
        factory = Factory.forProtocol(lambda: server)
        factory.protocolConnectionMade = None

        reactor = MemoryReactor()
        reactor.listenTCP(80, factory)

        t = FilePath(self.useFixture(TempDir()).join(b""))
        serviceaccount = t.child(b"serviceaccount")
        serviceaccount.makedirs()

        serviceaccount.child(b"ca.crt").setContent(_CA_CERT_PEM)
        serviceaccount.child(b"token").setContent(token)

        self.patch(
            os, "environ", {
                b"KUBERNETES_SERVICE_HOST": b"example.invalid.",
                b"KUBERNETES_SERVICE_PORT": b"443",
            },
        )

        agent = authenticate_with_serviceaccount(
            reactor, path=serviceaccount.path,
        )
        agent.request(b"GET", b"http://example.invalid.", headers)

        [(host, port, factory, _, _)] = reactor.tcpClients

        self.expectThat((host, port), Equals((b"example.invalid.", 80)))

        pump = ConnectionCompleter(reactor).succeedOnce()
        pump.pump()

        return server.data


    def test_bearer_token_authorization(self):
        """
        The ``IAgent`` returned adds an *Authorization* header to each request it
        issues.  The header includes the bearer token from the service account file.
        """
        token = bytes(uuid4())
        request_bytes = self._authorized_request(token=token, headers=None)

        # Sure would be nice to have an HTTP parser.
        self.assertThat(
            request_bytes,
            Contains(u"Authorization: Bearer {}".format(token).encode("ascii")),
        )


    def test_other_headers_preserved(self):
        """
        Other headers passed to the ``IAgent.request`` implementation are also
        sent in the request.
        """
        token = bytes(uuid4())
        headers = Headers({u"foo": [u"bar"]})
        request_bytes = self._authorized_request(token=token, headers=headers)
        self.expectThat(
            request_bytes,
            Contains(u"Authorization: Bearer {}".format(token).encode("ascii")),
        )
        self.expectThat(
            request_bytes,
            Contains(b"Foo: bar"),
        )


    def test_missing_ca_certificate(self):
        """
        If no CA certificate is found in the service account directory,
        ``authenticate_with_serviceaccount`` raises ``ValueError``.
        """
        t = FilePath(self.useFixture(TempDir()).join(b""))
        serviceaccount = t.child(b"serviceaccount")
        serviceaccount.makedirs()

        serviceaccount.child(b"ca.crt").setContent(b"not a cert pem")
        serviceaccount.child(b"token").setContent(b"token")

        self.patch(
            os, "environ", {
                b"KUBERNETES_SERVICE_HOST": b"example.invalid.",
                b"KUBERNETES_SERVICE_PORT": b"443",
            },
        )

        self.assertThat(
            lambda: authenticate_with_serviceaccount(
                MemoryReactor(), path=serviceaccount.path,
            ),
            raises(ValueError("No certificate authority certificate found.")),
        )


    def test_bad_ca_certificate(self):
        """
        If no CA certificate is found in the service account directory,
        ``authenticate_with_serviceaccount`` raises ``ValueError``.
        """
        t = FilePath(self.useFixture(TempDir()).join(b""))
        serviceaccount = t.child(b"serviceaccount")
        serviceaccount.makedirs()

        serviceaccount.child(b"ca.crt").setContent(
            b"-----BEGIN CERTIFICATE-----\n"
            b"not a cert pem\n"
            b"-----END CERTIFICATE-----\n"
        )
        serviceaccount.child(b"token").setContent(b"token")

        self.patch(
            os, "environ", {
                b"KUBERNETES_SERVICE_HOST": b"example.invalid.",
                b"KUBERNETES_SERVICE_PORT": b"443",
            },
        )

        self.assertThat(
            lambda: authenticate_with_serviceaccount(
                MemoryReactor(), path=serviceaccount.path,
            ),
            raises(ValueError(
                "Invalid certificate authority certificate found.",
                "[('PEM routines', 'PEM_read_bio', 'bad base64 decode')]",
            )),
        )



# Copyright Least Authority Enterprises.
# See LICENSE for details.

"""
A Kubernetes client.
"""

__all__ = [
    "version",
    "IObject", "IKubernetes", "IKubernetesClient",

    "KubernetesError", "UnrecognizedVersion", "UnrecognizedKind",
    "v1", "v1beta1", "iobject_from_raw", "iobject_to_raw",

    "memory_kubernetes",
    "network_kubernetes",  "network_kubernetes_from_context",
    "authenticate_with_serviceaccount",
    "authenticate_with_certificate",
]

from incremental import Version

from ._metadata import version_tuple as _version_tuple
version = __version__ = Version("txkube", *_version_tuple)

from ._exception import KubernetesError, UnrecognizedVersion, UnrecognizedKind
from ._interface import IObject, IKubernetes, IKubernetesClient

from ._model import (
    v1, v1beta1,
    iobject_from_raw,
    iobject_to_raw,
)

from ._authentication import (
    authenticate_with_serviceaccount, authenticate_with_certificate,
)
from ._network import network_kubernetes, network_kubernetes_from_context
from ._memory import memory_kubernetes



# Copyright Least Authority Enterprises.
# See LICENSE for details.

"""
Integration test generator for ``txkube.IKubernetesClient``.
"""

from operator import attrgetter, setitem
from functools import partial, wraps
from itertools import repeat

from zope.interface.verify import verifyObject

from eliot import Message, start_action
from eliot.twisted import DeferredContext

from testtools.matchers import (
    AnyMatch, MatchesAll, MatchesStructure, Is, IsInstance, Equals, Not,
    Contains, AfterPreprocessing, MatchesPredicate,
)

from testtools import run_test_with

from twisted.python.failure import Failure
from twisted.internet.defer import gatherResults
from twisted.internet.task import deferLater, cooperate
from twisted.web.http import NOT_FOUND, CONFLICT

from .._network import version_to_segments

from ..testing.matchers import MappingEquals
from ..testing import TestCase, AsynchronousDeferredRunTest

from .. import (
    KubernetesError,
    IKubernetesClient,
    v1, v1beta1,
    iobject_to_raw,
)

from .strategies import (
    labels, creatable_namespaces, configmaps,
    deployments, replicasets, pods, services,
)


def async(f):
    def _async(*a, **kw):
        # This is a pretty long timeout but it includes sitting around waiting
        # for Kubernetes to finish deleting things so we get better test
        # isolation...  Namespace deletion, specifically, can be slow, with
        # some attempts taking 5 minutes or more.
        kw["timeout"] = 600.0
        return AsynchronousDeferredRunTest(*a, **kw)
    return run_test_with(_async)(f)



def matches_metadata(expected):
    return MatchesStructure(
        metadata=MatchesStructure(
            namespace=Equals(expected.namespace),
            name=Equals(expected.name),
            # TODO: It would be nice to compare labels but currently that
            # results in an annoying failure because of the confusion of
            # representation of the empty value - {} vs None.
            # https://github.com/LeastAuthority/txkube/issues/66
            # labels=Equals(expected.labels),
        ),
    )


def matches_namespace(ns):
    return matches_metadata(ns.metadata)


def matches_configmap(configmap):
    return matches_metadata(configmap.metadata)


def matches_deployment(deployment):
    return MatchesAll(
        matches_metadata(deployment.metadata),
        # Augh.  I think some kind of "expected is None or values match"
        # matcher would help?
        MatchesStructure(
            spec=MatchesStructure(
                selector=Equals(deployment.spec.selector),
                template=matches_metadata(deployment.spec.template.metadata),
            ),
        ),
    )


def matches_replicaset(replicaset):
    return MatchesAll(
        matches_metadata(replicaset.metadata),
    )


def matches_pod(pod):
    return MatchesAll(
        matches_metadata(pod.metadata),
    )


def matches_service(service):
    return matches_metadata(service.metadata)


def has_uid():
    return MatchesStructure(
        metadata=MatchesStructure(
            uid=Not(Equals(None)),
        ),
    )


def is_active():
    return MatchesStructure(
        status=Equals(v1.NamespaceStatus.active()),
    )



def different_name(name):
    if len(name) > 1:
        return name[:-1]
    return name + u"x"



def _named(kind, name, namespace=None):
    return kind(metadata=v1.ObjectMeta(name=name, namespace=namespace))



def poll(reactor, action, f, intervals):
    """
    Repeatedly call a function.

    :param IReactorTime reactor: The reactor to use for scheduling.

    :param action: An Eliot action in the context of which to call ``f``.

    :param f: A no-argument callable which returns a ``Deferred``.

    :param interfaces: An iterable of numbers giving the delay to use between
        each call to ``f``.  This also limits the total number of calls.  When
        the iterable is exhausted, ``poll`` will stop calling ``f``.

    :return: A ``Deferred`` that fires with any errors encountered by ``f`` or
        with ``StopIteration`` if the end of ``intervals`` is reached.
    """
    intervals = iter(intervals)
    def g():
        with action.context():
            Message.log(poll_iteration=None)
            # Not bothering with DeferredContext here because when
            # deferLater calls g, it can restore the action context from
            # the closure (as it just did a couple lines above).
            d = f()
            d.addCallback(
                lambda ignored: deferLater(reactor, next(intervals), g),
            )
            return d
    d = g()
    d.addCallback(lambda ignored: g())
    return d


_TRY_AGAIN = (
    u"the object has been modified; "
    u"please apply your changes to the latest version and try again"
)

def _replace(client, obj, transformation, retries=2):
    """
    Try to replace the given object with one that differs by the given
    transformation.

    Try again if the server reports the object has changed.
    """
    d = client.replace(obj.transform(*transformation))
    def maybe_outdated(reason):
        reason.trap(KubernetesError)
        # Uugghh...  This seems to be the only way to identify the condition.
        if _TRY_AGAIN in reason.value.status.message:
            d = client.get(obj)
            d.addCallback(lambda latest: _replace(
                client, latest, transformation, retries - 1
            ))
            return d
        return reason
    if retries > 0:
        d.addErrback(maybe_outdated)
    return d



def does_not_exist(client, obj):
    """
    Wait for ``obj`` to not exist.

    :param IKubernetesClient client: The client to use to determine existence.

    :param IObject obj: The object the existence of which to check.

    :return: A ``Deferred`` that fires when the object does not exist.
    """
    action = start_action(action_type=u"poll:start", obj=iobject_to_raw(obj))
    with action.context():
        from twisted.internet import reactor
        d = poll(reactor, action, lambda: client.get(obj), repeat(0.5, 100))
        d = DeferredContext(d)
        # Once we get a NOT FOUND error, we're done.
        def trap_not_found(reason):
            reason.trap(KubernetesError)
            if reason.value.code == NOT_FOUND:
                return None
            return reason
        d.addErrback(trap_not_found)
        def trap_stop_iteration(reason):
            reason.trap(StopIteration)
            Message.log(does_not_exist=u"timeout", obj=iobject_to_raw(obj))
            return None
        d.addErrback(trap_stop_iteration)
        return d.addActionFinish()


def needs(**to_create):
    """
    Create a function decorator which will create certain Kubernetes objects
    before calling the decorated function and delete them after it completes.

    This requires the decorated functions accept a first argument with a
    ``client`` attribute bound to an ``IKubernetesClient``.

    :param to_create: Keyword arguments with values of Hypothesis strategies
        which build ``IObject`` providers.  After being created, these objects
        will be passed to the decorated function using the same keyword
        arguments.

    :return: A function decorator.
    """
    def build_objects():
        objs = {}
        names = set()
        for name, strategy in to_create.iteritems():
            while True:
                obj = strategy.example()
                if obj.metadata.name not in names:
                    names.add(obj.metadata.name)
                    break
            objs[name] = obj
        return objs
    return _needs_objects(build_objects)



def _needs_objects(get_objects):
    no_grace = v1.DeleteOptions(gracePeriodSeconds=0)

    def decorator(f):
        @wraps(f)
        def wrapper(self, *a, **kw):
            # Get the objects we're supposed to create.
            to_create = get_objects()

            # Check to make sure there aren't keyword argument conflicts.  I
            # doubt this is an exhaustive safety check.
            overlap = set(to_create) & set(kw)
            if overlap:
                raise TypeError(
                    "Conflict between @needs() and **kw: {}".format(overlap)
                )

            def cleanup(obj):
                # Delete an object, if it exists, and wait for it to stop
                # existing.
                d = self.client.delete(obj, no_grace)
                # Some Kubernetes objects take a while to be deleted.
                # Specifically, Namespaces go though a slow "Terminating"
                # phase where things they contain are cleaned up.  Delay
                # cleanup completion until objects are *really* gone.
                d.addBoth(lambda ignored: does_not_exist(self.client, obj))
                return d

            def create(obj):
                # Create some object and arrange for it to be deleted later.
                d = self.client.create(obj)
                def created(result):
                    self.addCleanup(lambda: cleanup(result))
                    return result
                d.addCallback(created)
                return d

            # Keep track of the objects we created - keyed on name.
            created = {}

            # Create an object, deleting an existing one that's in the way
            # first, if necessary.
            def ensure(obj):
                d = cleanup(obj)
                d.addCallback(lambda ignored: create(obj))
                return d

            # Create the objects.
            task = cooperate(
                ensure(obj).addCallback(partial(setitem, created, name))
                for (name, obj)
                in sorted(to_create.items())
            )
            d = task.whenDone()

            # Call the decorated function.
            d.addCallback(lambda ignored: kw.update(created))
            d.addCallback(lambda ignored: f(self, *a, **kw))
            return d
        return wrapper
    return decorator



class _NamespaceTestsMixin(object):
    @async
    @needs(obj=creatable_namespaces())
    def test_namespace(self, obj):
        """
        ``Namespace`` objects can be created and retrieved using the ``create``
        and ``list`` methods of ``IKubernetesClient``.
        """
        d = self.client.list(v1.Namespace)
        def check_namespaces(namespaces):
            self.assertThat(namespaces, IsInstance(v1.NamespaceList))
            # There are some built-in namespaces that we'll ignore.  If we
            # find the one we created, that's sufficient.
            self.assertThat(
                namespaces.items,
                AnyMatch(MatchesAll(matches_namespace(obj), has_uid(), is_active())),
            )
        d.addCallback(check_namespaces)
        return d


    @async
    @needs(obj=creatable_namespaces())
    def test_namespace_retrieval(self, obj):
        """
        A specific ``Namespace`` object can be retrieved by name using
        ``IKubernetesClient.get``.
        """
        return self._global_object_retrieval_by_name_test(
            obj,
            v1.Namespace,
            matches_namespace,
        )


    @async
    def test_namespace_replacement(self):
        """
        A specific ``Namespace`` object can be replaced by name using
        ``IKubernetesClient.replace``.
        """
        return self._object_replacement_test(
            None,
            creatable_namespaces(),
            matches_namespace,
        )


    @async
    @needs(obj=creatable_namespaces())
    def test_namespace_deletion(self, obj):
        """
        ``IKubernetesClient.delete`` can be used to delete ``Namespace``
        objects.
        """
        d = self.client.delete(obj)
        def deleted_namespace(ignored):
            return self.client.list(v1.Namespace)
        d.addCallback(deleted_namespace)
        def check_namespaces(collection):
            active = list(
                ns.metadata.name
                for ns
                in collection.items
                if ns.status.phase == u"Active"
            )
            self.assertThat(
                active,
                Not(Contains(obj.metadata.name)),
            )
        d.addCallback(check_namespaces)
        return d


    @async
    def test_duplicate_namespace_rejected(self):
        """
        ``IKubernetesClient.create`` returns a ``Deferred`` that fails with
        ``KubernetesError`` if it is called with a ``Namespace`` object
        with the same name as a *Namespace* which already exists.
        """
        return self._create_duplicate_rejected_test(
            None, creatable_namespaces(), u"namespaces", None,
        )



class _ConfigMapTestsMixin(TestCase):
    @async
    @needs(namespace=creatable_namespaces())
    def test_duplicate_configmap_rejected(self, namespace):
        """
        ``IKubernetesClient.create`` returns a ``Deferred`` that fails with
        ``KubernetesError`` if it is called with a ``ConfigMap`` object
        with the same name as a *ConfigMap* which already exists in the
        same namespace.
        """
        return self._create_duplicate_rejected_test(
            namespace, configmaps(), u"configmaps", None,
        )


    @async
    @needs(namespace=creatable_namespaces())
    def test_configmap(self, namespace):
        """
        ``ConfigMap`` objects can be created and retrieved using the ``create``
        and ``list`` methods of ``IKubernetesClient``.
        """
        self._create_list_test(
            namespace, configmaps(), v1.ConfigMap, v1.ConfigMapList,
            matches_configmap,
        )


    @async
    def test_configmap_retrieval(self):
        """
        A specific ``ConfigMap`` object can be retrieved by name using
        ``IKubernetesClient.get``.
        """
        return self._namespaced_object_retrieval_by_name_test(
            configmaps(),
            v1.ConfigMap,
            matches_configmap,
            group=None,
        )


    @async
    @needs(namespace=creatable_namespaces())
    def test_configmap_replacement(self, namespace):
        """
        A specific ``ConfigMap`` object can be replaced by name using
        ``IKubernetesClient.replace``.
        """
        return self._object_replacement_test(
            namespace,
            configmaps(),
            matches_configmap,
        )


    @async
    def test_configmap_deletion(self):
        """
        A specific ``ConfigMap`` object can be deleted by name using
        ``IKubernetesClient.delete``.
        """
        return self._namespaced_object_deletion_by_name_test(
            configmaps(),
            v1.ConfigMap,
        )


    ns_strategy = creatable_namespaces()
    @async
    @needs(ns_a=ns_strategy, ns_b=ns_strategy)
    def test_configmaps_sorted(self, ns_a, ns_b):
        """
        ``ConfigMap`` objects retrieved with ``IKubernetesClient.list`` appear in
        sorted order, with (namespace, name) as the sort key.
        """
        strategy = configmaps()
        objs = [
            strategy.example().transform(
                [u"metadata", u"namespace"], ns_a.metadata.name,
            ),
            strategy.example().transform(
                [u"metadata", u"namespace"], ns_b.metadata.name,
            ),
        ]
        d = gatherResults(list(self.client.create(obj) for obj in objs))
        def created_configmaps(ignored):
            return self.client.list(v1.ConfigMap)
        d.addCallback(created_configmaps)
        def check_configmaps(collection):
            self.expectThat(collection, items_are_sorted())
        d.addCallback(check_configmaps)
        return d



class _ReplicaSetTestsMixin(object):
    @async
    @needs(namespace=creatable_namespaces())
    def test_replicaset(self, namespace):
        """
        ``ReplicaSet`` objects can be created and retrieved using the ``create``
        and ``list`` methods of ``IKubernetesClient``.
        """
        return self._create_list_test(
            namespace, replicasets(), v1beta1.ReplicaSet,
            v1beta1.ReplicaSetList, matches_replicaset,
        )


    @async
    @needs(namespace=creatable_namespaces())
    def test_duplicate_replicaset_rejected(self, namespace):
        """
        ``IKubernetesClient.create`` returns a ``Deferred`` that fails with
        ``KubernetesError`` if it is called with a ``ReplicaSet`` object with
        the same name as a *ReplicaSet* which already exists in the same
        namespace.
        """
        return self._create_duplicate_rejected_test(
            namespace, replicasets(), u"replicasets", None,
        )


    @async
    def test_replicaset_retrieval(self):
        """
        A specific ``ReplicaSet`` object can be retrieved by name using
        ``IKubernetesClient.get``.
        """
        return self._namespaced_object_retrieval_by_name_test(
            replicasets(),
            v1beta1.ReplicaSet,
            matches_replicaset,
            group=None,
        )


    @async
    @needs(namespace=creatable_namespaces())
    def test_replicaset_replacement(self, namespace):
        """
        A specific ``ReplicaSet`` object can be replaced by name using
        ``IKubernetesClient.replace``.
        """
        return self._object_replacement_test(
            namespace,
            replicasets(),
            matches_replicaset,
        )


    @async
    def test_replicaset_deletion(self):
        """
        A specific ``ReplicaSet`` object can be deleted by name using
        ``IKubernetesClient.delete``.
        """
        return self._namespaced_object_deletion_by_name_test(
            replicasets(),
            v1beta1.ReplicaSet,
        )



class _PodTestsMixin(object):
    @async
    @needs(namespace=creatable_namespaces())
    def test_pod(self, namespace):
        """
        ``Pod`` objects can be created and retrieved using the ``create`` and
        ``list`` methods of ``IKubernetesClient``.
        """
        return self._create_list_test(
            namespace, pods(), v1.Pod,
            v1.PodList, matches_pod,
        )


    @async
    @needs(namespace=creatable_namespaces())
    def test_duplicate_pod_rejected(self, namespace):
        """
        ``IKubernetesClient.create`` returns a ``Deferred`` that fails with
        ``KubernetesError`` if it is called with a ``Pod`` object with the
        same name as a *Pod* which already exists in the same namespace.
        """
        return self._create_duplicate_rejected_test(
            namespace, pods(), u"pods", None,
        )


    @async
    def test_pod_retrieval(self):
        """
        A specific ``Pod`` object can be retrieved by name using
        ``IKubernetesClient.get``.
        """
        return self._namespaced_object_retrieval_by_name_test(
            pods(),
            v1.Pod,
            matches_pod,
            group=None,
        )


    @async
    @needs(namespace=creatable_namespaces())
    def test_pod_replacement(self, namespace):
        """
        A specific ``Pod`` object can be replaced by name using
        ``IKubernetesClient.replace``.
        """
        return self._object_replacement_test(
            namespace,
            pods(),
            matches_pod,
        )


    @async
    def test_pod_deletion(self):
        """
        A specific ``Pod`` object can be deleted by name using
        ``IKubernetesClient.delete``.
        """
        return self._namespaced_object_deletion_by_name_test(
            pods(),
            v1.Pod,
        )



class _DeploymentTestsMixin(object):
    @async
    @needs(namespace=creatable_namespaces())
    def test_deployment(self, namespace):
        """
        ``Deployment`` objects can be created and retrieved using the ``create``
        and ``list`` methods of ``IKubernetesClient``.
        """
        return self._create_list_test(
            namespace, deployments(), v1beta1.Deployment,
            v1beta1.DeploymentList, matches_deployment,
        )


    @async
    @needs(namespace=creatable_namespaces())
    def test_duplicate_deployment_rejected(self, namespace):
        """
        ``IKubernetesClient.create`` returns a ``Deferred`` that fails with
        ``KubernetesError`` if it is called with a ``Deployment`` object with
        the same name as a *Deployment* which already exists in the same
        namespace.
        """
        return self._create_duplicate_rejected_test(
            namespace, deployments(), u"deployments", u"extensions",
        )


    @async
    def test_deployment_retrieval(self):
        """
        A specific ``Deployment`` object can be retrieved by name using
        ``IKubernetesClient.get``.
        """
        return self._namespaced_object_retrieval_by_name_test(
            deployments(),
            v1beta1.Deployment,
            matches_deployment,
            group=u"extensions",
        )


    @async
    @needs(namespace=creatable_namespaces())
    def test_deployment_replacement(self, namespace):
        """
        A specific ``Deployment`` object can be replaced by name using
        ``IKubernetesClient.replace``.
        """
        return self._object_replacement_test(
            namespace,
            deployments(),
            matches_deployment,
        )


    @async
    def test_deployment_deletion(self):
        """
        A specific ``Deployment`` object can be deleted by name using
        ``IKubernetesClient.delete``.
        """
        return self._namespaced_object_deletion_by_name_test(
            deployments(),
            v1beta1.Deployment,
        )



class _ServiceTestsMixin(object):
    @async
    @needs(namespace=creatable_namespaces())
    def test_service(self, namespace):
        """
        ``Service`` objects can be created and retrieved using the ``create`` and
        ``list`` methods of ``IKubernetesClient``.
        """
        self._create_list_test(
            namespace, services(), v1.Service, v1.ServiceList,
            matches_service,
        )


    @async
    @needs(namespace=creatable_namespaces())
    def test_duplicate_service_rejected(self, namespace):
        """
        ``IKubernetesClient.create`` returns a ``Deferred`` that fails with
        ``KubernetesError`` if it is called with a ``ConfigMap`` object
        with the same name as a *ConfigMap* which already exists in the
        same namespace.
        """
        return self._create_duplicate_rejected_test(
            namespace, configmaps(), u"configmaps", None,
        )


    @async
    def test_service_retrieval(self):
        """
        A specific ``Service`` object can be retrieved by name using
        ``IKubernetesClient.get``.
        """
        return self._namespaced_object_retrieval_by_name_test(
            services(),
            v1.Service,
            matches_service,
            group=None,
        )


    @async
    @needs(namespace=creatable_namespaces())
    def test_service_replacement(self, namespace):
        """
        A specific ``Service`` object can be replaced by name using
        ``IKubernetesClient.replace``.
        """
        return self._object_replacement_test(
            namespace,
            services(),
            matches_service,
        )


    @async
    def test_service_deletion(self):
        """
        A specific ``Service`` object can be deleted by name using
        ``IKubernetesClient.delete``.
        """
        return self._namespaced_object_deletion_by_name_test(
            services(),
            v1.Service,
        )



def kubernetes_client_tests(get_kubernetes):
    class KubernetesClientIntegrationTests(
        _NamespaceTestsMixin,
        _ConfigMapTestsMixin,
        _DeploymentTestsMixin,
        _ReplicaSetTestsMixin,
        _PodTestsMixin,
        _ServiceTestsMixin,
        TestCase
    ):
        def setUp(self):
            super(KubernetesClientIntegrationTests, self).setUp()
            self.kubernetes = get_kubernetes(self)
            self.client = self.kubernetes.client()
            self.addCleanup(self._cleanup)


        def _cleanup(self):
            pool = getattr(self.client.agent, "_pool", None)
            if pool is None:
                return None
            from twisted.internet import reactor
            return gatherResults([
                pool.closeCachedConnections(),
                # Semi-work-around for
                # https://twistedmatrix.com/trac/ticket/8998
                deferLater(reactor, 1.0, lambda: None),
            ])


        def test_interfaces(self):
            """
            The client provides ``txkube.IKubernetesClient``.
            """
            verifyObject(IKubernetesClient, self.client)


        @async
        def test_not_found(self):
            """
            ``IKubernetesClient.list`` returns a ``Deferred`` that fails with
            ``KubernetesError`` when the server responds with an HTTP NOT
            FOUND status.
            """
            # Invent a type that the server isn't going to recognize.  This
            # could happen if we're talking to a server that is missing some
            # extension we thought it had, for example.
            class Mythical(object):
                apiVersion = u"v6txkube2"
                kind = u"Mythical"
                metadata = v1.ObjectMeta()

            # The client won't know where to route the request without this.
            # There is a more general problem here that the model is missing
            # some unpredictable information about where the relevant APIs are
            # exposed in the URL hierarchy.
            version_to_segments[Mythical.apiVersion] = (
                u"apis", u"extensions", Mythical.apiVersion,
            )
            self.addCleanup(lambda: version_to_segments.pop(Mythical.apiVersion))

            d = self.client.list(Mythical)
            def failed(reason):
                self.assertThat(reason, IsInstance(Failure))
                reason.trap(KubernetesError)
                self.assertThat(
                    reason.value,
                    MatchesStructure(
                        code=Equals(NOT_FOUND),
                        status=Equals(v1.Status(
                            metadata={},
                            status=u"Failure",
                            message=u"the server could not find the requested resource",
                            reason=u"NotFound",
                            details=dict(),
                            code=NOT_FOUND,
                        )),
                    ),
                )
            d.addBoth(failed)
            return d


        def _global_object_retrieval_by_name_test(self, obj, kind, matches):
            """
            Verify that a particular kind of non-namespaced Kubernetes object (such as
            *Namespace* or *PersistentVolume*) can be retrieved by name by
            calling ``IKubernetesClient.get`` with the ``IObject``
            corresponding to that kind as long as the object has its *name*
            metadata populated.
            """
            d = self.client.get(_named(kind, name=obj.metadata.name))
            def got_object(retrieved):
                self.assertThat(retrieved, matches(obj))
            d.addCallback(got_object)
            return d


        def _object_replacement_test(self, namespace, strategy, matches):
            """
            Verify that an existing object can be replaced by a new one.

            :param namespace: If the object belongs in a namespace, the name
                of an existing namespace to create it in.  ``None`` if not.

            :param strategy: A Hypothesis strategy for building the object to
                create and the object to replace it with.

            :param matches: A one-argument caller which takes the expected
                object and returns a testtools matcher for it.  Since created
                objects have some unpredictable server-generated fields, this
                matcher can compare just the important, predictable parts of
                the object.

            :return: A ``Deferred`` that fires when the behavior has been
                verified.
            """
            original = strategy.example()
            if namespace is not None:
                original = original.transform(
                    [u"metadata", u"namespace"], namespace.metadata.name,
                )
            # Different kinds of objects have different constraints on what
            # they can be replaced with.  To avoid running into those, make
            # the replacement just like the original but with some different
            # labels.  Hopefully this will always be allowed.
            #
            # Also, make sure _some_ labels are set to avoid triggering extra
            # server processing in some cases.  For example, Deployment
            # inherits labels from its template spec if its own labels are
            # empty.
            replacement_labels = labels().filter(lambda labels: len(labels)).example()

            d = self.client.create(original)
            def created(created):
                # Use the created object as the basis for the replacement
                # object.  Various server-populated fields (varying between
                # different kinds of objects) must match in the replacement.
                # Using our original object as the starting point will mess
                # these.
                transformation = (
                    [u"metadata", u"labels"],
                    replacement_labels,
                )
                return _replace(self.client, created, transformation, retries=1)
            d.addCallback(created)
            def replaced(result):
                self.expectThat(
                    result.metadata.labels,
                    MappingEquals(replacement_labels),
                )
                return self.client.get(original)
            d.addCallback(replaced)
            def retrieved(result):
                self.expectThat(
                    result.metadata.labels,
                    MappingEquals(replacement_labels),
                )
            d.addCallback(retrieved)
            return d


        def _create_list_test(self, namespace, strategy, cls, list_cls, matches):
            """
            Verify that a particular kind of namespaced Kubernetes object (such as
            *Service* or *Secret*) can be created using
            ``IKubernetesClient.create`` and then appears in the result of
            ``IKubernetesClient.list``.
            """
            # Get an object in the namespace.
            expected = strategy.example().transform(
                [u"metadata", u"namespace"],
                namespace.metadata.name,
            )
            d = self.client.create(expected)
            def created(actual):
                self.assertThat(actual, matches(expected))
                return self.client.list(cls)
            d.addCallback(created)
            def check(collection):
                self.assertThat(collection, IsInstance(list_cls))
                self.assertThat(collection.items, AnyMatch(matches(expected)))
            d.addCallback(check)
            return d


        def _create_duplicate_rejected_test(self, namespace, strategy, kind, group):
            """
            Verify an object cannot be created if its name (and maybe namespace) is
            already taken.

            :param IObject obj: Some object to create and try to create again.

            :param unicode kind: The name of the collection of the *kind* of
                ``obj``, lowercase.  For example, *configmaps*.
            """
            if group is None:
                kind_identifier = kind
            else:
                kind_identifier = u"{}.{}".format(kind, group)

            obj = strategy.example()
            # Put it in the namespace.
            if namespace is not None:
                obj = obj.transform(
                    [u"metadata", u"namespace"],
                    namespace.metadata.name,
                )
            d = self.client.create(obj)
            def created(ignored):
                return self.client.create(obj)
            d.addCallback(created)
            def failed(reason):
                self.assertThat(reason, IsInstance(Failure))
                reason.trap(KubernetesError)
                self.assertThat(
                    reason.value,
                    MatchesStructure(
                        code=Equals(CONFLICT),
                        status=Equals(v1.Status(
                            metadata={},
                            status=u"Failure",
                            # XXX This message is a little janky.  "namespaces
                            # ... already exists"?  How about "Namespace
                            # ... already exists"?  Maybe report it to
                            # Kubernetes.
                            message=u"{} \"{}\" already exists".format(
                                kind_identifier, obj.metadata.name,
                            ),
                            reason=u"AlreadyExists",
                            details=dict(
                                name=obj.metadata.name,
                                kind=kind,
                                group=group,
                            ),
                            code=CONFLICT,
                        )),
                    ),
                )
            d.addBoth(failed)
            return d


        ns_strategy = creatable_namespaces()
        @needs(
            victim_namespace=ns_strategy,
            bystander_namespace=ns_strategy,
        )
        def _namespaced_object_deletion_by_name_test(
            self, strategy, cls, victim_namespace, bystander_namespace
        ):
            """
            Verify that a particular kind of namespaced Kubernetes object (such as
            *Deployment* or *Service*) can be deleted by name by calling
            ``IKubernetesClient.delete`` with the ``IObject`` corresponding to
            that kind as long as the object has its *name* and *namespace*
            metadata populated.

            :param strategy: A Hypothesis strategy for building the namespaced
                object to create and then retrieve.

            :param cls: The ``IObject`` implementation corresponding to the
                objects ``strategy`` can build.

            :param v1.Namespace victim_namespace: An existing namespace into
                which the probe object can be created (not part of the
                interface; value supplied by the decorator).

            :param v1.Namespace bystander_namespace: An existing namespace
                into which another object can be created (not part of the
                interface; value supplied by the decorator).

            :return: A ``Deferred`` that fires when the behavior has been
                verified.
            """
            # This object will be the target of the deletion.
            victim = strategy.example()
            # These bystanders should be left alone.
            bystander_a = strategy.example()
            bystander_b = strategy.example()
            victim = victim.transform(
                # Put the victim in the victim namespace.
                [u"metadata", u"namespace"], victim_namespace.metadata.name,
            )
            bystander_a = bystander_a.transform(
                # Put a bystander in the victim namespace, too.
                [u"metadata", u"namespace"], victim_namespace.metadata.name,
                # Also, be completely certain the bystander has a different
                # name.
                [u"metadata", u"name"], different_name(victim.metadata.name),
            )
            bystander_b = bystander_b.transform(
                # Put a bystander in another namespace too.
                [u"metadata", u"namespace"], bystander_namespace.metadata.name,
            )
            d = gatherResults(list(
                self.client.create(o)
                for o
                in [victim, bystander_a, bystander_b]
            ))
            def created_object(ignored):
                return self.client.delete(
                    _named(
                        cls,
                        namespace=victim.metadata.namespace, name=victim.metadata.name,
                    ),
                    v1.DeleteOptions(gracePeriodSeconds=0),
                )
            d.addCallback(created_object)
            def deleted_object(result):
                self.expectThat(result, Is(None))
                return self.client.list(cls)
            d.addCallback(deleted_object)
            def listed_objects(collection):
                def key(obj):
                    return (obj.metadata.name, obj.metadata.namespace)
                obj_names = set(
                    key(obj)
                    for obj
                    in collection.items
                )
                self.expectThat(
                    obj_names,
                    MatchesAll(
                        Contains(key(bystander_a)),
                        Contains(key(bystander_b)),
                        Not(Contains(key(victim))),
                    ),
                )
            d.addCallback(listed_objects)
            return d


        @needs(namespace=creatable_namespaces())
        def _namespaced_object_retrieval_by_name_test(self, strategy, cls, matches, group, namespace):
            """
            Verify that a particular kind of namespaced Kubernetes object (such as
            *ConfigMap* or *PersistentVolumeClaim*) can be retrieved by name
            by calling ``IKubernetesClient.get`` with the ``IObject``
            corresponding to that kind as long as the object has its *name*
            and *namespace* metadata populated.

            :param strategy: A Hypothesis strategy for building the namespaced
                object to create and then retrieve.

            :param cls: The ``IObject`` implementation corresponding to the
                objects ``strategy`` can build.

            :param matches: A one-argument caller which takes the expected
                object and returns a testtools matcher for it.  Since created
                objects have some unpredictable server-generated fields, this
                matcher can compare just the important, predictable parts of
                the object.

            :param unicode group: The name of the API group responsible for
                this retrieval.  Objects are arbitrarily collected into such
                groupings.  Look at the Kubernetes API Operations
                documentation to find out what group contains the retrieval
                API for the type under test.

            :param v1.Namespace namespace: An existing namespace into which
                the probe object can be created.

            :return: A ``Deferred`` that fires when the behavior has been
                verified.
            """
            kind = cls.kind.lower()
            obj = strategy.example()
            # Move it to the namespace for this test.
            obj = obj.transform([u"metadata", u"namespace"], namespace.metadata.name)
            d = self.client.create(obj)
            def created_object(created):
                return self.client.get(_named(
                    cls,
                    namespace=obj.metadata.namespace, name=obj.metadata.name,
                ))
            d.addCallback(created_object)
            def got_object(retrieved):
                self.expectThat(retrieved, matches(obj))
                # Try retrieving an object with the same name but a different
                # namespace.  We shouldn't find it.
                #
                # First, compute a legal but non-existing namespace name.
                bogus_namespace = different_name(obj.metadata.namespace)
                return self.client.get(
                    _named(
                        cls,
                        namespace=bogus_namespace,
                        name=obj.metadata.name,
                    ),
                )
            d.addCallback(got_object)
            def check_error(reason):
                self.assertThat(reason, IsInstance(Failure))
                reason.trap(KubernetesError)
                if group is None:
                    fmt = u"{kind} \"{name}\" not found"
                else:
                    fmt = u"{kind}.{group} \"{name}\" not found"
                details = dict(
                    kind=u"{}s".format(kind),
                    name=obj.metadata.name,
                    group=group,
                )
                self.assertThat(
                    reason.value,
                    MatchesStructure(
                        code=Equals(NOT_FOUND),
                        status=Equals(v1.Status(
                            metadata={},
                            status=u"Failure",
                            message=fmt.format(**details),
                            reason=u"NotFound",
                            details=details,
                            code=NOT_FOUND,
                        )),
                    ),
                )
            d.addBoth(check_error)
            return d

    return KubernetesClientIntegrationTests



def items_are_sorted():
    """
    Match an ObjectCollection if its items can be iterated in the Kubernetes
    canonical sort order - lexicographical by namespace, name.
    """
    def key(obj):
        return (
            getattr(obj.metadata, "namespace", None),
            obj.metadata.name,
        )

    def is_sorted(items, key):
        return list(items) == sorted(items, key=key)

    return AfterPreprocessing(
        attrgetter("items"),
        MatchesPredicate(
            partial(is_sorted, key=key),
            u"%s is not sorted by namespace, name",
        ),
    )



# Copyright Least Authority Enterprises.
# See LICENSE for details.

"""
A Kubernetes client which uses Twisted to interact with Kubernetes
via HTTP.
"""

from os.path import expanduser
from json import loads, dumps

from zope.interface import implementer

import attr
from attr import validators

from pem import parse

from twisted.python.reflect import namedAny
from twisted.python.failure import Failure
from twisted.python.url import URL
from twisted.python.filepath import FilePath

from twisted.internet.defer import succeed

from twisted.web.iweb import IBodyProducer, IAgent
from twisted.web.http import OK, CREATED
from twisted.web.client import Agent, readBody

from eliot import Message, start_action
from eliot.twisted import DeferredContext

from pykube import KubeConfig

from . import (
    IObject, IKubernetes, IKubernetesClient, KubernetesError,
    iobject_from_raw, iobject_to_raw,
    authenticate_with_certificate,
)

def network_kubernetes(**kw):
    """
    Create a new ``IKubernetes`` provider which can be used to create clients.

    :param twisted.python.url.URL base_url: The root of the Kubernetes HTTPS
        API to interact with.

    :param twisted.web.iweb.IAgent agent: An HTTP agent to use to issue
        requests.  Defaults to a new ``twisted.web.client.Agent`` instance.
        See ``txkube.authenticate_with_serviceaccount`` and
        ``txkube.authenticate_with_certificate`` for helpers for creating
        agents that interact well with Kubernetes servers.

    :return IKubernetes: The Kubernetes service.
    """
    return _NetworkKubernetes(**kw)



def network_kubernetes_from_context(reactor, context, path=None):
    """
    Create a new ``IKubernetes`` provider based on a kube config file.

    :param reactor: A Twisted reactor which will be used for I/O and
        scheduling.

    :param unicode context: The name of the kube config context from which to
        load configuration details.

    :param FilePath path: The location of the kube config file to use.

    :return IKubernetes: The Kubernetes service described by the named
        context.
    """
    if path is None:
        path = FilePath(expanduser(u"~/.kube/config"))

    config = KubeConfig.from_file(path.path)
    context = config.contexts[context]
    cluster = config.clusters[context[u"cluster"]]
    user = config.users[context[u"user"]]

    base_url = URL.fromText(cluster[u"server"].decode("ascii"))
    [ca_cert] = parse(cluster[u"certificate-authority"].bytes())

    [client_cert] = parse(user[u"client-certificate"].bytes())
    [client_key] = parse(user[u"client-key"].bytes())

    agent = authenticate_with_certificate(
        reactor, base_url, client_cert, client_key, ca_cert,
    )

    return network_kubernetes(
        base_url=base_url,
        agent=agent,
    )


# It would be simpler to use FileBodyProducer(BytesIO(...)) but:
#
#  - https://twistedmatrix.com/trac/ticket/9003
#  - https://github.com/twisted/treq/issues/161
@implementer(IBodyProducer)
@attr.s(frozen=True)
class _BytesProducer(object):
    _data = attr.ib(validator=validators.instance_of(bytes), repr=False)

    @property
    def length(self):
        return len(self._data)

    def startProducing(self, consumer):
        consumer.write(self._data)
        return succeed(None)

    def stopProducing(self):
        pass

    def pauseProducing(self):
        pass

    def resumeProducing(self):
        pass



@implementer(IKubernetesClient)
@attr.s(frozen=True)
class _NetworkClient(object):
    _apiVersion = u"v1"

    kubernetes = attr.ib(validator=validators.provides(IKubernetes))
    agent = attr.ib(validator=validators.provides(IAgent))

    def _request(self, method, url, headers=None, bodyProducer=None):
        action = start_action(
            action_type=u"network-client:request",
            method=method,
            url=url.asText(),
        )
        with action.context():
            d = self.agent.request(
                method, url.asText().encode("ascii"), headers, bodyProducer,
            )
            return DeferredContext(d).addActionFinish()


    def _get(self, url):
        return self._request(b"GET", url)


    def _delete(self, url, options):
        bodyProducer = None
        if options is not None:
            bodyProducer = _BytesProducer(dumps(iobject_to_raw(options)))
        return self._request(b"DELETE", url, bodyProducer=bodyProducer)


    def _post(self, url, obj):
        return self._request(
            b"POST", url, bodyProducer=_BytesProducer(dumps(obj)),
        )


    def _put(self, url, obj):
        return self._request(
            b"PUT", url, bodyProducer=_BytesProducer(dumps(obj)),
        )


    def create(self, obj):
        """
        Issue a I{POST} to create the given object.
        """
        action = start_action(
            action_type=u"network-client:create",
        )
        with action.context():
            url = self.kubernetes.base_url.child(*collection_location(obj))
            document = iobject_to_raw(obj)
            Message.log(submitted_object=document)
            d = DeferredContext(self._post(url, document))
            d.addCallback(check_status, (CREATED,))
            d.addCallback(readBody)
            d.addCallback(loads)
            d.addCallback(log_response_object, action)
            d.addCallback(iobject_from_raw)
            return d.addActionFinish()


    def replace(self, obj):
        """
        Issue a I{PUT} to replace an existing object with a new one.
        """
        action = start_action(
            action_type=u"network-client:replace",
        )
        with action.context():
            url = self.kubernetes.base_url.child(*object_location(obj))
            document = iobject_to_raw(obj)
            Message.log(submitted_object=document)
            d = DeferredContext(self._put(url, document))
            d.addCallback(check_status, (OK,))
            d.addCallback(readBody)
            d.addCallback(loads)
            d.addCallback(log_response_object, action)
            d.addCallback(iobject_from_raw)
            return d.addActionFinish()


    def get(self, obj):
        """
        Issue a I{GET} to retrieve the given object.

        The object must have identifying metadata such as a namespace and a
        name but other fields are ignored.
        """
        action = start_action(
            action_type=u"network-client:get",
            kind=obj.kind,
            name=obj.metadata.name,
            namespace=getattr(obj.metadata, "namespace", None),
        )
        with action.context():
            url = self.kubernetes.base_url.child(*object_location(obj))
            d = DeferredContext(self._get(url))
            d.addCallback(check_status, (OK,))
            d.addCallback(readBody)
            d.addCallback(loads)
            d.addCallback(log_response_object, action)
            d.addCallback(iobject_from_raw)
            return d.addActionFinish()


    def delete(self, obj, options=None):
        """
        Issue a I{DELETE} to delete the given object.

        :param v1.DeleteOptions options: Optional details to control some
            consequences of the deletion.
        """
        action = start_action(
            action_type=u"network-client:delete",
            kind=obj.kind,
            name=obj.metadata.name,
            namespace=getattr(obj.metadata, "namespace", None),
        )
        with action.context():
            url = self.kubernetes.base_url.child(*object_location(obj))
            d = DeferredContext(self._delete(url, options))
            d.addCallback(check_status, (OK,))
            d.addCallback(readBody)
            d.addCallback(lambda raw: None)
            return d.addActionFinish()


    def list(self, kind):
        """
        Issue a I{GET} to retrieve objects of a given kind.
        """
        action = start_action(
            action_type=u"network-client:list",
            kind=kind.kind,
            apiVersion=kind.apiVersion,
        )
        with action.context():
            url = self.kubernetes.base_url.child(*collection_location(kind))
            d = DeferredContext(self._get(url))
            d.addCallback(check_status, (OK,))
            d.addCallback(readBody)
            d.addCallback(lambda body: iobject_from_raw(loads(body)))
            return d.addActionFinish()



def object_location(obj):
    """
    Get the URL for a specific object.

    :param IObject obj: The object the URL for which to get.

    :return tuple[unicode]: Some path segments to stick on to a base URL top
        construct the location for the given object.
    """
    return collection_location(obj) + (obj.metadata.name,)



version_to_segments = {
    u"v1": (u"api", u"v1"),
    u"v1beta1": (u"apis", u"extensions", u"v1beta1"),
}


def collection_location(obj):
    """
    Get the URL for the collection of objects like ``obj``.

    :param obj: Either a type representing a Kubernetes object kind or an
        instance of such a type.

    :return tuple[unicode]: Some path segments to stick on to a base URL to
        construct the location of the collection of objects like the one
        given.
    """
    # TODO kind is not part of IObjectLoader and we should really be loading
    # apiVersion off of this object too.
    kind = obj.kind
    apiVersion = obj.apiVersion

    prefix = version_to_segments[apiVersion]

    collection = kind.lower() + u"s"

    if IObject.providedBy(obj):
        # Actual objects *could* have a namespace...
        namespace = obj.metadata.namespace
    else:
        # Types representing a kind couldn't possible.
        namespace = None

    if namespace is None:
        # If there's no namespace, look in the un-namespaced area.
        return prefix + (collection,)

    # If there is, great, look there.
    return prefix + (u"namespaces", namespace, collection)



@implementer(IKubernetes)
@attr.s(frozen=True)
class _NetworkKubernetes(object):
    """
    ``_NetworkKubernetes`` knows the location of a particular
    Kubernetes deployment and gives out clients which speak to that
    deployment.
    """
    base_url = attr.ib(validator=validators.instance_of(URL))
    _agent = attr.ib(
        default=attr.Factory(lambda: Agent(namedAny("twisted.internet.reactor"))),
    )

    def client(self):
        return _NetworkClient(self, self._agent)



def log_response_object(document, action):
    """
    Emit an Eliot log event belonging to the given action describing the given
    response.

    :param document: Anything Eliot loggable (but presumably a parsed
        Kubernetes response document).

    :param action: The Eliot action to which to attach the event.

    :return: ``document``
    """
    action.add_success_fields(response_object=document)
    return document


def check_status(response, expected):
    if response.code not in expected:
        d = KubernetesError.from_response(response)
        d.addCallback(Failure)
        return d
    return response



# Copyright Least Authority Enterprises.
# See LICENSE for details.

"""
Tests for ``txkube._invariants``.
"""

from zope.interface import Interface, implementer

from pyrsistent import InvariantException, PClass, field

from ..testing import TestCase
from .._invariants import instance_of, provider_of


class IDummy(Interface):
    pass



@implementer(IDummy)
class Dummy(PClass):
    an_int = field(invariant=instance_of(int))
    an_interface = field(invariant=provider_of(IDummy))



class InstanceOfTests(TestCase):
    """
    Tests for ``instance_of``.
    """
    def test_valid(self):
        self.assertEqual(3, Dummy(an_int=3).an_int)


    def test_invalid(self):
        self.assertRaises(
            InvariantException,
            lambda: Dummy(an_int=b"bytes"),
        )




class ProviderOfTests(TestCase):
    """
    Tests for ``provider_of``.
    """
    def test_valid(self):
        d = Dummy()
        self.assertIs(d, Dummy(an_interface=d).an_interface)


    def test_invalid(self):
        self.assertRaises(
            InvariantException,
            lambda: Dummy(an_interface=b"bytes"),
        )



# Copyright Least Authority Enterprises.
# See LICENSE for details.

"""
An interface to Swagger specifications.

This module provides the ability to automatically derive somewhat sensible
Python type definitions from definitions in a Swagger specification.  The
Python types enforce the constraints defined by the Swagger specification as
best they can.  They also support being serialized down to raw (JSON-y)
objects and loaded from such objects.
"""

from json import load
from datetime import datetime
from itertools import chain

from dateutil.parser import parse as parse_iso8601

from zope.interface import Attribute, Interface, implementer

from pyrsistent import (
    CheckedValueTypeError, PClass, PVector, pvector, field, pvector_field,
    pmap_field, freeze, pmap,
)

from twisted.python.compat import nativeString
from twisted.python.reflect import fullyQualifiedName


class NotClassLike(Exception):
    """
    An attempt was made to treat a Swagger definition as though it were
    compatible with Python classes but it is not (for example, it is a simple
    type like *string* or *integer*).
    """



class Swagger(PClass):
    """
    A ``Swagger`` contains a single Swagger specification.

    Public attributes of this class correspond to the top-level properties in
    a `Swagger specification <http://swagger.io/>`_.

    :ivar dict _pclasses: A cache of the `PClass`\ es that have been
        constructed for definitions from this specification already.  This
        allows multiple requests for the same definition to be satisfied with
        the same type object (rather than a new type object which has all the
        same attributes and behavior).  This plays better with Python's type
        system than the alternative.
    """
    info = field()
    paths = field()
    definitions = field(factory=freeze)
    securityDefinitions = field()
    security = field()
    swagger = field()

    _pclasses = field(mandatory=True, type=dict)

    @classmethod
    def from_path(cls, spec_path):
        """
        Load a specification from a path.

        :param FilePath spec_path: The location of the specification to read.
        """
        with spec_path.open() as spec_file:
            return cls.from_document(load(spec_file))


    @classmethod
    def from_document(cls, document):
        """
        Load a specification from some Python objects.

        :param dict document: An object like the one that might be created by
            parsing a Swagger JSON specification string.
        """
        return cls(_pclasses={}, **document)


    def pclass_for_definition(self, name, constant_fields=pmap()):
        """
        Get a ``pyrsistent.PClass`` subclass representing the Swagger definition
        in this specification which corresponds to the given name.

        :param unicode name: The name of the definition to use.

        :return: A Python class which can be used to represent the Swagger
            definition of the given name.
        """
        try:
            cls = self._pclasses[name]
        except KeyError:
            definition = self.definitions[name]
            kind = self._identify_kind(definition)
            if kind is None:
                raise NotClassLike(name, definition)
            generator =  getattr(self, "_model_for_{}".format(kind))
            model = generator(name, definition, constant_fields)
            cls = model.pclass()
            self._pclasses[name] = cls
        return cls


    def _identify_kind(self, definition):
        """
        Determine what kind of thing the given definition seems to resemble.

        This must be inferred from the structure of the definition.  For
        example, if it includes the *properties* then the thing is sort of
        like a Python class.

        :param pyrsistent.PMap definition: A Swagger definition to categorize.
            This will be a value like the one found at
            ``spec["definitions"][name]``.

        :return: ``"CLASS"`` for things that are class-like.  ``None``
            otherwise (though it would be good to extend this).
        """
        if u"properties" in definition:
            return "CLASS"
        return None


    def _model_for_CLASS(self, name, definition, constant_fields):
        """
        Model a Swagger definition that is like a Python class.

        :param unicode name: The name of the definition from the
            specification.

        :param pyrsistent.PMap definition: A Swagger definition to categorize.
            This will be a value like the one found at
            ``spec["definitions"][name]``.

        :param dict constant_fields: Additional fields to set as constants on
            the resulting class.  These will override properties from the
            Swagger definition if they collide.
        """
        return _ClassModel.from_swagger(
            self.pclass_for_definition, name, definition,
            constant_fields,
        )



class IRangeModel(Interface):
    """
    An ``IRangeModel`` provider models the range of values a type can take on.
    """
    def pyrsistent_invariant():
        """
        :return: A pyrsistent invariant which enforces the range.  It accepts one
            argument (a value to validate) and returns a two-tuple.  The first
            element is ``True`` if the value is in the range, ``False`` otherwise.
            If it is ``False``, the second element gives a human-readable
            description of how the value fell out of the range.
        """


class ITypeModel(Interface):
    """
    An ``ITypeModel`` provider models the type of a value.

    Such types could be associated with the value of a property or a value
    found in an array or mapping.
    """
    python_types = Attribute("tuple of python types compatible with this type")
    factory = Attribute("An optional callable for converting values to this type.")

    def pclass_field_for_type(required):
        """
        Create a pyrsistent field for this model object for use with a PClass.

        :param bool required: Whether the field should be mandatory or
            optional.

        :return: A pyrsistent field descriptor for an attribute which is
            usable with values of this type.
        """



@implementer(ITypeModel)
class _BasicTypeModel(PClass):
    """
    A ``_BasicTypeModel`` represents a type composed (roughly) of a "single"
    value.

    Specifically, this is used for Swagger types of *boolean*, *integer*,
    *string* (except *date-time*), and references to other *object*\ s.

    :ivar tuple(type) python_types: The Python types which correspond to this
        the modeled Swagger type.

    :ivar IRangeModel range: Optionally, a model of the range of values
        allowed (necessary because, for example, a Python integer type can
        hold arbitrarily large values but the modeled type may be limited to
        the range representable in 32 bits).

    :ivar factory: A one-argument callable which can convert non-canonical
        representations to canonical representations (and which will be
        applied to values provided for use with this type).
    """
    python_types = field(mandatory=True)
    range = field(mandatory=True, initial=None)
    factory = field(mandatory=True, initial=None)
    serializer = field(mandatory=True, initial=lambda: lambda format, value: value)

    def _pyrsistent_invariant(self, required):
        """
        Create a pyrsistent invariant which reflects this model.

        :param bool required: ``True`` if the invariant will require a
            non-``None`` value, ``False`` if ``None`` is allowed in addition
            to whatever ``self.range`` describes.

        :return: A callable suitable for use as a pyrsistent invariant.
        """
        if self.range is None:
            return None
        invariant = self.range.pyrsistent_invariant()
        if required:
            return invariant
        def optional(v):
            if v is None:
                return (True, u"")
            return invariant(v)
        return optional


    def _pyrsistent_factory(self, required):
        """
        Create a pyrsistent field factory which reflects this model.

        :param bool required: ``True`` if the factory will require a
            non-``None`` value, ``False`` if ``None`` is allowed to bypass
            ``self.factory``.

        :return: A callable suitable for use as a pyrsistent factory.
        """
        if self.factory is None:
            return None
        if required:
            return self.factory
        def optional(v):
            if v is None:
                return None
            if isinstance(v, self.python_types):
                return v
            return self.factory(v)
        return optional


    def _pyrsistent_serializer(self, required):
        """
        Create a pyrsistent serialization function.

        :param bool required: ``True`` if this value must appear in the
            serialized form.  ``False`` if it may be omitted when it has a
            ``None`` value.

        :return: A callable suitable for use as a pyrsistent field serializer.
        """
        if required:
            return self.serializer

        def serialize(format, value):
            if value is None:
                return omit
            return self.serializer(format, value)
        return serialize


    def pclass_field_for_type(self, required):
        """
        Construct a pyrsistent field reflecting this model.

        The field uses the model's invariant and factory.  If ``required`` is
        ``False``, an initial value of ``None`` is also supplied.

        :return: The field descriptor.
        """
        extra = {}
        python_types = self.python_types

        invariant = self._pyrsistent_invariant(required)
        if invariant is not None:
            extra[u"invariant"] = invariant

        if not required:
            python_types += (type(None),)
            extra[u"initial"] = None

        factory = self._pyrsistent_factory(required)
        if factory is not None:
            extra[u"factory"] = factory

        extra[u"serializer"] = self._pyrsistent_serializer(required)

        return field(
            mandatory=required, type=python_types, **extra
        )



def provider_invariant(interface):
    """
    :param zope.interface.Interface interface: An interface to require.

    :return: A pyrsistent invariant which requires that values provide the
        given interface.
    """
    return lambda o: (
        interface.providedBy(o),
        "does not provide {}".format(fullyQualifiedName(interface)),
    )



def itypemodel_field():
    """
    :return: A pyrsistent field for an attribute which much refer to an
        ``ITypeModel`` provider.
    """
    return field(invariant=provider_invariant(ITypeModel))



@implementer(ITypeModel)
class _ArrayTypeModel(PClass):
    """
    An ``_ArrayTypeModel`` represents a type which is a homogeneous array.

    Specifically, this is used for the Swagger type *array*.

    :ivar tuple(type) python_types: The Python types which correspond to this
        the modeled Swagger type.

    :ivar ITypeModel element_type: The type model that applies to elements of
        the array.
    """
    element_type = itypemodel_field()

    @property
    def python_types(self):
        # Cheat a bit and make pyrsistent synthesize a type for us...
        # Amusingly, it's a regular set internally so also freeze it so it's
        # okay to put it back in to field again.
        return freeze(self.pclass_field_for_type(True).type)


    def pclass_field_for_type(self, required):
        # XXX ignores the range's pyrsistent_invariant
        return pvector_field(self.element_type.python_types, optional=not required)



# TODO It might make more sense to handle this the same way as a reference to
# another object.  The current implementation makes it tricky because it's not
# straightforward to name these inline/nested objects (as it is
# straightforward to name a top-level definition).
@implementer(ITypeModel)
class _MappingTypeModel(PClass):
    """
    A ``_MappingTypeModel`` represents a type which is a homogeneous mapping
    with ``unicode`` keys.

    Specifically, this is used for the Swagger type *object* when it occurs
    inline in another definition.

    :ivar ITypeModel value_field: The type model that applies to values in the
        mapping.
    """
    value_type = itypemodel_field()

    def pclass_field_for_type(self, required):
        # XXX ignores the range's pyrsistent_invariant
        return pmap_field(
            key_type=unicode,
            value_type=self.value_type.python_types,
            optional=not required,
        )



class _AttributeModel(PClass):
    """
    An ``_AttributeModel`` models an attribute of a class-like definition.

    Specifically, this is used for an item in a Swagger *object*\ s
    *properties*.

    :ivar unicode name: The name of the attribute.

    :ivar unicode descriptor: The human-meaningful description of the
        attribute.

    :ivar itypemodel_field type_model: The type model for values this
        attribute can reference.

    :ivar bool required: ``True`` if a value conforming to ``type_model`` must
        be supplied.  ``False`` if ``None`` is allowed in addition to values
        allowed by ``type_model``.
    """
    name = field(type=unicode)
    description = field(type=unicode)
    type_model = itypemodel_field()
    required = field(type=bool)

    def pclass_field_for_attribute(self):
        """
        :return: A pyrsistent field reflecting this attribute and its type model.
        """
        return self.type_model.pclass_field_for_type(required=self.required)



class _ConstantModel(PClass):
    """
    A ``_ConstantModel`` models a constant valued attribute.

    This does not correspond to any part of a Swagger specification.  Instead,
    it's part of how a Swagger definition is usefully exposed to Python.  It
    can be used to override certain fields from a Swagger definition (for
    example, when there is a known constant value which the Swagger
    specification can't formally represent).

    :ivar unicode name: The name of the attribute.
    :ivar value: The value of the attribute.
    """
    name = field(type=unicode)
    value = field()

    def pclass_field_for_attribute(self):
        """
        :return: Exactly this model's value, creating a constant-valued attribute
        on a class it is used with.
        """
        return self.value



@implementer(IRangeModel)
class _IntegerRange(PClass):
    """
    ``_IntegerRange`` represents a contiguous range of integer values between
    two boundaries.

    :ivar int min: The lower bound, inclusive.
    :ivar int max: The upper bound, inclusive.
    """
    min = field(type=(int, long))
    max = field(type=(int, long))


    @classmethod
    def from_unsigned_bits(cls, n):
        """
        Create a range corresponding to that of an unsigned integer of ``n``
        bits.
        """
        return cls(min=0, max=2 ** n - 1)

    @classmethod
    def from_signed_bits(cls, n):
        """
        Create a range corresponding to that of a two's complement signed integer
        of ``n`` bits.
        """
        m = n - 1
        return cls(min=-2 ** m, max=2 ** m - 1)


    def pyrsistent_invariant(self):
        """
        :return: An invariant which rejects values outside of *(min, max)*.
        """
        return lambda v: (
            self.min <= v <= self.max,
            "{!r} out of required range ({}, {})".format(v, self.min, self.max),
        )



def _parse_iso8601(text):
    """
    Maybe parse an ISO8601 datetime string into a datetime.

    :param text: Either a ``unicode`` string to parse or any other object
        (ideally a ``datetime`` instance) to pass through.

    :return: A ``datetime.datetime`` representing ``text``.  Or ``text`` if it
        was anything but a ``unicode`` string.
    """
    if isinstance(text, unicode):
        try:
            return parse_iso8601(text)
        except ValueError:
            raise CheckedValueTypeError(
                None, (datetime,), unicode, text,
            )
    # Let pyrsistent reject it down the line.
    return text



def _isoformat(format, v):
    return v.isoformat().decode("ascii")



class _ClassModel(PClass):
    """
    A ``_ClassModel`` represents a type with a number of named, heterogeneous
    fields - something like a Python class.

    Specifically, this is used to represent top-level Swagger definitions of
    type *object*.

    :cvar _basic_types: A mapping from the simpler Swagger types to
        ``ITypeModel`` providers which represent them.

    :ivar unicode name: The name of the definition.
    :ivar unicode doc: The description of the definition.
    :ivar PVector attributes: Models for the fields of the definition.
    """
    _basic_types = {
        (u"boolean", None): _BasicTypeModel(python_types=(bool,)),
        (u"integer", u"int32"): _BasicTypeModel(
            # XXX Kubernetes uses this to mean unsigned 32 bit integer.
            # Swagger spec says it is for signed 32 bit integer.  Since we're
            # trying to *use* Kubernetes ...
            python_types=(int, long), range=_IntegerRange.from_unsigned_bits(32),
        ),
        (u"integer", u"int64"): _BasicTypeModel(
            python_types=(int, long), range=_IntegerRange.from_unsigned_bits(64),
        ),
        (u"string", None): _BasicTypeModel(python_types=(unicode,)),
        (u"string", u"byte"): _BasicTypeModel(python_types=(bytes,)),
        (u"string", u"date-time"): _BasicTypeModel(
            python_types=(datetime,),
            factory=_parse_iso8601,
            serializer=_isoformat,
        ),
        (u"string", u"int-or-string"): _BasicTypeModel(
            python_types=(unicode, int, long),
        ),
    }

    name = field(type=unicode)
    doc = field(type=unicode)
    attributes = field(type=PVector, factory=pvector)

    @classmethod
    def _type_model_for_spec(cls, pclass_for_definition, spec):
        if spec.get(u"type") == u"array":
            # "array" type definitions represent an array of some other thing.
            # Get a model for whatever the nested thing is and put it into an
            # array model.
            element_type = cls._type_model_for_spec(
                pclass_for_definition, spec[u"items"],
            )
            return _ArrayTypeModel(element_type=element_type)

        if spec.get(u"type") == u"object":
            # "object" type definitions represent a mapping from unicode to
            # some other thing.  Get a model for whatever the values are
            # supposed to be and put that into a mapping model.
            value_type = cls._type_model_for_spec(
                pclass_for_definition, spec[u"additionalProperties"],
            )
            return _MappingTypeModel(value_type=value_type)

        if u"$ref" in spec:
            # "$ref" type definitions refer to some other definition in the
            # specification.  Look that up and use it.
            name = spec[u"$ref"]
            assert name.startswith(u"#/definitions/")
            name = name[len(u"#/definitions/"):]
            try:
                # For anything that's class-like (basically, has properties)
                # we'll get another PClass for the reference target.
                python_type = pclass_for_definition(name)
            except NotClassLike as e:
                # For anything that's not class-like (it could be a simple
                # type like a string), create a simple type model for it instead.
                #
                # The spec is conveniently available in the exception
                # arguments.  This is an awful way to pass information around.
                # Probably some refactoring is in order.
                _, spec = e.args
                return cls._type_model_for_spec(
                    pclass_for_definition, spec,
                )
            else:
                # For our purposes, the pclass we got is just another basic
                # type we can model.
                return _BasicTypeModel(
                    python_types=(python_type,),
                    factory=python_type.create,
                    serializer=lambda format, value: value.serialize(format),
                )

        # If it wasn't any of those kinds of things, maybe it's just a simple
        # type.  Look up the corresponding model object in the static mapping.
        return cls._basic_types[spec[u"type"], spec.get(u"format", None)]


    @classmethod
    def _attribute_for_property(
        cls, pclass_for_definition, name, required, definition
    ):
        type_model = cls._type_model_for_spec(pclass_for_definition, definition)
        return _AttributeModel(name=name, required=required, type_model=type_model)


    @classmethod
    def _attributes_for_definition(cls, pclass_for_definition, definition):
        required = definition.get(u"required", [])
        for prop, spec in definition[u"properties"].items():
            yield cls._attribute_for_property(
                pclass_for_definition, prop, prop in required, spec,
            )


    @classmethod
    def from_swagger(cls, pclass_for_definition, name, definition, constant_fields):
        """
        Create a new ``_ClassModel`` from a single Swagger definition.

        :param pclass_for_definition: A callable like
            ``Swagger.pclass_for_definition`` which can be used to resolve
            type references encountered in the definition.

        :param unicode name: The name of the definition.

        :param definition: The Swagger definition to model.  This will be a
            value like the one found at ``spec["definitions"][name]``.

        :param constant_fields: A mapping of properties overrides.

        :return: A new model for the given definition.
        """
        return cls(
            name=name,
            doc=definition.get(u"description", name),
            attributes=chain((
                attr
                for attr
                in cls._attributes_for_definition(pclass_for_definition, definition)
                if constant_fields is None or attr.name not in constant_fields
            ), (
                _ConstantModel(name=name, value=value)
                for (name, value)
                in constant_fields.items()
            )),
        )


    def pclass(self):
        """
        Create a ``pyrsistent.PClass`` subclass representing this class.
        """
        content = {
            attr.name: attr.pclass_field_for_attribute()
            for attr
            in self.attributes
        }
        content["__doc__"] = nativeString(self.doc)
        content["serialize"] = _serialize_with_omit
        return type(nativeString(self.name), (PClass,), content)


omit = object()
def _serialize_with_omit(self, format=None):
    return {
        key: value
        for (key, value)
        in PClass.serialize(self, format).iteritems()
        if value is not omit
    }



class INameTranslator(Interface):
    """
    An ``INameTranslator`` translates from a name convenient for use in Python
    to a name used in a Swagger definition.
    """
    def translate(name):
        """
        Translate the name from Python to Swagger.

        :param unicode name: The Python name.
        :return unicode: The Swagger name.
        """



@implementer(INameTranslator)
class IdentityTranslator(object):
    """
    ``IdentityTranslator`` provides the identity translation.  In other words,
    it is a no-op.
    """
    def translate(self, name):
        return name



@implementer(INameTranslator)
class UsePrefix(PClass):
    """
    ``UsePrefix`` provides a translation which prepends a prefix.  This is
    useful, for example, for a versioning convention where many definition
    names have a prefix like *v1*.

    :ivar unicode prefix: The prefix to prepend.
    """
    prefix = field(mandatory=True, type=unicode)

    def translate(self, name):
        return self.prefix + name



class PClasses(PClass):
    """
    ``PClasses`` provides a somewhat easier to use interface to PClasses
    representing definitions from a Swagger definition.  For example::

    .. code-block: python

       spec = Swagger.from_path(...)
       v1beta1 = PClasses(
           specification=spec,
           name_translator=UsePrefix(u"v1beta1."),
       )
       class Deployment(v1beta1[u"Deployment"]):
           ...
    """
    specification = field(mandatory=True, type=Swagger)
    name_translator = field(
        mandatory=True, initial=IdentityTranslator(),
        invariant=provider_invariant(INameTranslator),
    )

    def __getitem__(self, name):
        """
        Get a PClass for the translation of the given name.

        :param unicode name: The Python name to translate and look up.

        :return: A ``PClass`` subclass for the identified Swagger definition.
        """
        name = self.name_translator.translate(name)
        return self.specification.pclass_for_definition(name)



class VersionedPClasses(object):
    """
    ``VersionedPClasses`` provides a somewhat easier to use interface to
    PClasses representing definitions from a Swagger definition.  For
    example::

    .. code-block: python

       spec = Swagger.from_path(...)
       v1beta1 = VersionedPClasses(
           spec, u"v1beta1", u"kind", u"apiVersion",
       )
       deployment = v1beta1.Deployment(...)
    """
    def __init__(self, spec, version, name_field=None, version_field=None):
        self.spec = spec
        self.version = version
        self.name_field = name_field
        self.version_field = version_field


    def __getattr__(self, name):
        name = name.decode("ascii")
        constant_fields = {}
        if self.name_field is not None:
            constant_fields[self.name_field] = name
        if self.version_field is not None:
            constant_fields[self.version_field] = self.version
        definition_name = self.version + u"." + name
        try:
            return self.spec.pclass_for_definition(
                definition_name, constant_fields=constant_fields,
            )
        except KeyError:
            raise AttributeError(name)



# Copyright Least Authority Enterprises.
# See LICENSE for details.

"""
Tests for ``txkube.testing._eliot``.
"""

from eliot import start_action, add_destination, remove_destination

from testtools.matchers import Contains

from .. import TestCase

from .._eliot import _eliottree


class EliotTreeTests(TestCase):
    """
    Tests for ``txkube.testing._eliot._eliottree``.
    """
    def test_tree(self):
        """
        ``_eliottree`` returns a ``bytes`` string containing a rendered tree of
        Eliot actions and messages.
        """
        events = []
        add_destination(events.append)
        self.addCleanup(lambda: remove_destination(events.append))

        with start_action(action_type=u"foo"):
            pass

        # I don't know exactly what the tree rendering looks like.  That's why
        # I'm using eliot-tree!  So this assertion is sort of lame.
        self.assertThat(
            _eliottree(events).decode("utf-8"),
            Contains(u"foo@1/started"),
        )



# Copyright Least Authority Enterprises.
# See LICENSE for details.

from json import loads

from twisted.web.http import NOT_FOUND, CONFLICT
from twisted.web.client import readBody


def _full_kind(details):
    """
    Determine the full kind (including a group if applicable) for some failure
    details.

    :see: ``v1.Status.details``
    """
    kind = details[u"kind"]
    if details.get(u"group") is not None:
        kind += u"." + details[u"group"]
    return kind



class KubernetesError(Exception):
    """
    Kubernetes has returned an error for some attempted operation.

    :ivar int code: The HTTP response code.
    :ivar Status status: The *v1.Status* returned in the response.
    """
    def __init__(self, code, status):
        self.code = code
        self.status = status

    @classmethod
    def not_found(cls, details):
        # Circular imports :(  See below.
        from ._model import v1
        kind = _full_kind(details)
        return cls(
            NOT_FOUND,
            v1.Status(
                status=u"Failure",
                message=u'{kind} "{name}" not found'.format(
                    kind=kind, name=details[u"name"],
                ),
                reason=u"NotFound",
                details=details,
                metadata={},
                code=NOT_FOUND,
            ),
        )

    @classmethod
    def already_exists(cls, details):
        # Circular imports :(  See below.
        from ._model import v1
        kind = _full_kind(details)
        return cls(
            CONFLICT,
            v1.Status(
                status=u"Failure",
                message=u'{kind} "{name}" already exists'.format(
                    kind=kind, name=details[u"name"],
                ),
                reason=u"AlreadyExists",
                details=details,
                metadata={},
                code=CONFLICT,
            ),
        )

    @classmethod
    def object_modified(cls, details):
        from ._model import v1
        kind = _full_kind(details)
        fmt = (
            u'Operation cannot be fulfilled on {kind} "{name}": '
            u'the object has been modified; '
            u'please apply your changes to the latest version and try again'
        )
        return cls(
            CONFLICT,
            v1.Status(
                code=CONFLICT,
                details=details,
                message=fmt.format(kind=kind, name=details[u"name"]),
                metadata={},
                reason=u'Conflict',
                status=u'Failure',
            ),
        )


    @classmethod
    def from_response(cls, response):
        """
        Create a ``KubernetesError`` for the given error response from a
        Kubernetes server.

        :param twisted.web.iweb.IResponse response: The response to inspect
            for the error details.

        :return Deferred(KubernetesError): The error with details attached.
        """
        d = readBody(response)
        # txkube -> _exception -> _model -> txkube :(
        #
        # Stick the import here to break the cycle.
        #
        # This is usually what happens with the expose-it-through-__init__
        # style, I guess.
        from ._model import iobject_from_raw
        d.addCallback(lambda body: cls(response.code, iobject_from_raw(loads(body))))
        return d


    def __repr__(self):
        return "<KubernetesError: code = {}; status = {}>".format(
            self.code, self.status,
        )

    __str__ = __repr__



class UnrecognizedVersion(ValueError):
    """
    An object *apiVersion* was encountered that we don't know about.

    :ivar unicode apiVersion: The API version encountered.
    :ivar object obj: The whole marshalled object.
    """
    def __init__(self, apiVersion, obj):
        ValueError.__init__(self, apiVersion)
        self.apiVersion = apiVersion
        self.obj = obj



class UnrecognizedKind(ValueError):
    """
    An object *kind* was encountered that we don't know about.

    :ivar unicode apiVersion: The API version encountered.
    :ivar unicode kind: The object kind encountered.
    :ivar object obj: The whole marshalled object.
    """
    def __init__(self, apiVersion, kind, obj):
        ValueError.__init__(self, apiVersion, kind)
        self.apiVersion = apiVersion
        self.kind = kind
        self.obj = obj



# Copyright Least Authority Enterprises.
# See LICENSE for details.

"""
testtools matchers for txkube.
"""

import operator

import attr

from pyrsistent import PClass, field

from testtools.matchers import Mismatch


class MappingLikeEquals(PClass):
    expected = field()

    comparator = operator.eq
    mismatch_string = "!="

    def __new__(cls, expected):
        # Provide positional argument support.
        # https://github.com/tobgu/pyrsistent/issues/94
        return super(MappingLikeEquals, cls).__new__(cls, expected=expected)


    def fields(self, obj):
        raise NotImplementedError()


    def get_field(self, obj, field):
        raise NotImplementedError()


    def __str__(self):
        return "%s(%r)" % (self.__class__.__name__, self.expected)


    def match(self, actual):
        if self.comparator(actual, self.expected):
            return None
        return _MappingLikeMismatch(
            self.fields, self.get_field, actual, self.mismatch_string,
            self.expected,
        )



class MappingEquals(MappingLikeEquals):
    def fields(self, obj):
        return obj.keys()


    def get_field(self, obj, key):
        return obj[key]



class AttrsEquals(MappingLikeEquals):
    def fields(self, obj):
        return list(field.name for field in attr.fields(type(obj)))


    def get_field(self, obj, key):
        return getattr(obj, key)



class PClassEquals(MappingLikeEquals):
    def fields(self, obj):
        return obj._pclass_fields.keys()


    def get_field(self, obj, key):
        try:
            return getattr(obj, key)
        except AttributeError:
            raise KeyError(key)



class _Missing(object):
    def __repr__(self):
        return "<<missing>>"

_missing = _Missing()



class _MappingLikeMismatch(Mismatch):
    def __init__(self, get_fields, get_field, actual, mismatch_string, reference,
                 reference_on_right=True):
        self._get_fields = get_fields
        self._get_field = get_field
        self._actual = actual
        self._mismatch_string = mismatch_string
        self._reference = reference
        self._reference_on_right = reference_on_right


    def describe(self):
        if type(self._actual) != type(self._reference):
            return (
                "type mismatch:\n"
                "reference = %s (%s)\n"
                "actual    = %s (%s)\n"
            ) % (
                type(self._reference), self._reference,
                type(self._actual), self._actual,
            )

        mismatched = []
        fields = self._get_fields(self._actual)
        for a_field in sorted(fields):
            try:
                actual = self._get_field(self._actual, a_field)
            except KeyError:
                actual = _missing
            try:
                reference = self._get_field(self._reference, a_field)
            except KeyError:
                reference = _missing
            if actual != reference:
                mismatched.append(dict(
                    field=a_field,
                    actual=actual,
                    reference=reference,
                ))

        extra = set(self._get_fields(self._reference)) - set(fields)
        for a_field in sorted(extra):
            reference = self._get_field(self._reference, a_field)
            mismatched.append(dict(
                field=a_field,
                reference=reference,
                actual=_missing,
            ))

        return "field mismatch:\n" + "".join(
            "field: %(field)s\n"
            "reference = %(reference)s\n"
            "actual    = %(actual)s\n" % mismatch
            for mismatch
            in mismatched
        )



# Copyright Least Authority Enterprises.
# See LICENSE for details.

"""
Tests for ``txkube.network_kubernetes``.

See ``get_kubernetes`` for pre-requisites.
"""

from os import environ

from zope.interface import implementer
from zope.interface.verify import verifyClass

from testtools.matchers import AnyMatch, ContainsDict, Equals

from eliot.testing import capture_logging

from twisted.test.proto_helpers import MemoryReactor
from twisted.python.url import URL
from twisted.web.client import Agent

from ..testing import TestCase
from ..testing.integration import kubernetes_client_tests

from .. import (
    IObject, v1, network_kubernetes, network_kubernetes_from_context,
)

from .._network import collection_location


def get_kubernetes(case):
    """
    Create a real ``IKubernetes`` provider, taking necessary
    configuration details from the environment.

    To use this set ``TXKUBE_INTEGRATION_CONTEXT`` to a context in your
    ``kubectl`` configuration.  Corresponding details about connecting to a
    cluster will be loaded from that configuration.
    """
    try:
        context = environ["TXKUBE_INTEGRATION_CONTEXT"]
    except KeyError:
        case.skipTest("Cannot find TXKUBE_INTEGRATION_CONTEXT in environment.")
    else:
        from twisted.internet import reactor
        return network_kubernetes_from_context(reactor, context)


class KubernetesClientIntegrationTests(kubernetes_client_tests(get_kubernetes)):
    """
    Integration tests which interact with a network-accessible
    Kubernetes deployment via ``txkube.network_kubernetes``.
    """



class CollectionLocationTests(TestCase):
    """
    Tests for ``collection_location``.
    """
    def _test_collection_location(self, version, kind, expected, namespace, instance):
        """
        Verify that ``collection_location`` for a particular version, kind,
        namespace, and Python object.

        :param unicode version: The *apiVersion* of the object to test.
        :param unicode kind: The *kind* of the object to test.

        :param tuple[unicode] expected: A representation of the path of the
            URL which should be produced.

        :param namespace: The namespace the Python object is to claim - a
            ``unicode`` string or ``None``.

        :param bool instance: Whether to make the Python object an instance
            (``True``) or a class (``False``)..
        """
        k = kind
        n = namespace
        @implementer(IObject)
        class Mythical(object):
            apiVersion = version
            kind = k

            metadata = v1.ObjectMeta(namespace=n)

            def serialize(self):
                return {}

        verifyClass(IObject, Mythical)

        if instance:
            o = Mythical()
        else:
            o = Mythical

        self.assertThat(
            collection_location(o),
            Equals(expected),
        )


    def test_v1_type(self):
        """
        ``collection_location`` returns a tuple representing an URL path like
        */api/v1/<kind>s* when called with an ``IObject`` implementation of a
        *v1* Kubernetes object kind.
        """
        self._test_collection_location(
            u"v1", u"Mythical", (u"api", u"v1", u"mythicals"),
            namespace=None,
            instance=False,
        )


    def test_v1_instance(self):
        """
        ``collection_location`` returns a tuple representing an URL path like
        */api/v1/namespace/<namespace>/<kind>s* when called with an
        ``IObject`` provider representing Kubernetes object of a *v1* kind.
        """
        self._test_collection_location(
            u"v1", u"Mythical",
            (u"api", u"v1", u"namespaces", u"ns", u"mythicals"),
            namespace=u"ns",
            instance=True,
        )


    def test_v1beta1_type(self):
        """
        ``collection_location`` returns a tuple representing an URL path like
        */apis/extensions/v1beta1/<kind>s* when called with an ``IObject``
        implementation of a *v1beta1* Kubernetes object kind.
        """
        self._test_collection_location(
            u"v1beta1", u"Mythical",
            (u"apis", u"extensions", u"v1beta1", u"mythicals"),
            namespace=None,
            instance=False,
        )


    def test_v1beta1_instance(self):
        """
        ``collection_location`` returns a tuple representing an URL path like
        */apis/extensions/v1beta1/<kind>s* when called with an ``IObject``
        implementation of a *v1beta1* Kubernetes object kind.
        """
        self._test_collection_location(
            u"v1beta1", u"Mythical",
            (u"apis", u"extensions", u"v1beta1", u"namespaces", u"ns",
             u"mythicals"),
            namespace=u"ns",
            instance=True,
        )


class ExtraNetworkClientTests(TestCase):
    """
    Direct tests for ``_NetworkClient`` that go beyond the guarantees of
    ``IKubernetesClient``.
    """
    @capture_logging(
        lambda self, logger: self.expectThat(
            logger.messages,
            AnyMatch(ContainsDict({
                u"action_type": Equals(u"network-client:list"),
                u"apiVersion": Equals(u"v1"),
                u"kind": Equals(u"Pod"),
            })),
        ),
    )
    def test_list_logging(self, logger):
        """
        ``_NetworkClient.list`` logs an Eliot event describing its given type.
        """
        client = network_kubernetes(
            base_url=URL.fromText(u"http://127.0.0.1/"),
            agent=Agent(MemoryReactor()),
        ).client()
        client.list(v1.Pod)



# Copyright Least Authority Enterprises.
# See LICENSE for details.

"""
xUnit TestCase for txkube testing.
"""

from fixtures import CompoundFixture

from testtools import TestCase as TesttoolsTestCase
from testtools.twistedsupport import AsynchronousDeferredRunTest

from ._eliot import CaptureEliotLogs


class AsynchronousDeferredRunTest(AsynchronousDeferredRunTest):
    """
    An asynchronous runner supporting Eliot.
    """
    def _get_log_fixture(self):
        """
        Add ``CaptureEliotLogs`` to the log fixtures which receive special
        treatment so as to be "cleaned up" in the timeout case.

        This ensures eliot logs are reported when tests time out - something
        that will not happen using the normal ``useFixture`` API.

        See <https://bugs.launchpad.net/testtools/+bug/897196>.
        """
        return CompoundFixture([
            super(AsynchronousDeferredRunTest, self)._get_log_fixture(),
            CaptureEliotLogs(),
        ])



class TestCase(TesttoolsTestCase):
    """
    A base class for test cases which automatically uses the
    ``CaptureEliotLogs`` fixture.
    """
    # expectThat and Hypothesis don't communicate well about when the
    # test has failed.  Give them a little help.  These two Hypothesis
    # hooks will check for a flag that testtools sets when it thinks
    # the test has failed and turn it into something Hypothesis can
    # recognize.
    def setup_example(self):
        try:
            # TesttoolsTestCase starts without this attribute set at all.  Get
            # us back to that state.  It won't be set at all on the first
            # setup_example call, nor if the previous run didn't have a failed
            # expectation.
            del self.force_failure
        except AttributeError:
            pass

    def teardown_example(self, ignored):
        if getattr(self, "force_failure", False):
            # If testtools things it's time to stop, translate this into a
            # test failure exception that Hypothesis can see.  This lets
            # Hypothesis know when it has found a falsifying example.  Without
            # it, Hypothesis can't see which of its example runs caused
            # problems.
            self.fail("expectation failed")



# Copyright Least Authority Enterprises.
# See LICENSE for details.

"""
Tests for ``txkube.testing.matchers``.
"""

import attr

from pyrsistent import PClass, field

from testtools.matchers import Is, Equals

from .. import TestCase
from ..matchers import MappingEquals, AttrsEquals, PClassEquals


class MappingEqualsTests(TestCase):
    """
    Tests for ``MappingEquals``.
    """
    def test_equals(self):
        """
        ``MappingEquals.match`` returns ``None`` when comparing two ``dict`` which
        compare equal with ``==``.
        """
        self.assertThat(
            MappingEquals({u"foo": u"bar"}).match({u"foo": u"bar"}),
            Is(None),
        )


    def test_mismatch(self):
        """
        ``MappingEquals.match`` returns a mismatch when comparing two ``dict``
        which do not compare equal with ``==``.
        """
        # Same keys, different value.
        mismatch = MappingEquals({u"foo": u"bar"}).match({u"foo": u"baz"})
        self.expectThat(
            mismatch.describe(),
            Equals(
                u"field mismatch:\n"
                u"field: foo\n"
                u"reference = bar\n"
                u"actual    = baz\n"
            ),
        )

        # Different types altogether.
        mismatch = MappingEquals(0).match({0: 1})
        self.expectThat(
            mismatch.describe(),
            Equals(
                u"type mismatch:\n"
                u"reference = <type 'int'> (0)\n"
                u"actual    = <type 'dict'> ({0: 1})\n"
            ),
        )

        # Actual value missing a key.
        mismatch = MappingEquals({u"foo": u"bar"}).match({})
        self.expectThat(
            mismatch.describe(),
            Equals(
                u"field mismatch:\n"
                u"field: foo\n"
                u"reference = bar\n"
                u"actual    = <<missing>>\n"
            ),
        )

        # Expected value missing a key.
        mismatch = MappingEquals({}).match({u"foo": u"baz"})
        self.expectThat(
            mismatch.describe(),
            Equals(
                u"field mismatch:\n"
                u"field: foo\n"
                u"reference = <<missing>>\n"
                u"actual    = baz\n"
            ),
        )

        # The matcher has a nice string representation.
        self.expectThat(
            str(MappingEquals({})),
            Equals("MappingEquals({})"),
        )



class AttrsEqualsTests(TestCase):
    """
    Tests for ``AttrsEquals``.
    """
    @attr.s
    class attrs(object):
        foo = attr.ib()


    def test_equals(self):
        """
        ``AttrsEquals.match`` returns ``None`` when comparing two attrs-based
        instances which compare equal with ``==``.
        """
        self.assertThat(
            AttrsEquals(self.attrs(u"bar")).match(self.attrs(u"bar")),
            Is(None),
        )


    def test_mismatch(self):
        """
        ``AttrsEquals.match`` returns a mismatch when comparing two attrs-based
        instances which do not compare equal with ``==``.
        """
        # Different value for the single attribute.
        mismatch = AttrsEquals(self.attrs(u"bar")).match(self.attrs(u"baz"))
        self.expectThat(
            mismatch.describe(),
            Equals(
                u"field mismatch:\n"
                u"field: foo\n"
                u"reference = bar\n"
                u"actual    = baz\n"
            ),
        )

        # Different types altogether.
        mismatch = AttrsEquals(self.attrs(0)).match(1)
        self.expectThat(
            mismatch.describe(),
            Equals(
                u"type mismatch:\n"
                u"reference = <class 'txkube.testing.test.test_matchers.attrs'> (attrs(foo=0))\n"
                u"actual    = <type 'int'> (1)\n"
            ),
        )

        # The matcher has a nice string representation.
        self.expectThat(
            str(AttrsEquals(self.attrs(u"bar"))),
            Equals("AttrsEquals(attrs(foo=u'bar'))"),
        )



class PClassEqualsTests(TestCase):
    """
    Tests for ``PClassEquals``.
    """
    class pclass(PClass):
        foo = field()
        bar = field()


    def test_equals(self):
        """
        ``PClassEquals.match`` returns ``None`` when comparing two PClass-based
        instances which compare equal with ``==``.
        """
        self.assertThat(
            PClassEquals(self.pclass(foo=u"bar")).match(self.pclass(foo=u"bar")),
            Is(None),
        )


    def test_mismatch(self):
        """
        ``PClassEquals.match`` returns a mismatch when comparing two ``dict``
        which do not compare equal with ``==``.
        """
        # Same attributes, different value.
        mismatch = PClassEquals(self.pclass(foo=u"bar")).match(self.pclass(foo=u"baz"))
        self.expectThat(
            mismatch.describe(),
            Equals(
                u"field mismatch:\n"
                u"field: foo\n"
                u"reference = bar\n"
                u"actual    = baz\n"
            ),
        )

        # Different types altogether.
        mismatch = PClassEquals(0).match(self.pclass(foo=1))
        self.expectThat(
            mismatch.describe(),
            Equals(
                u"type mismatch:\n"
                u"reference = <type 'int'> (0)\n"
                u"actual    = <class 'txkube.testing.test.test_matchers.pclass'> (pclass(foo=1))\n"
            ),
        )

        # Actual value missing an attribute.
        mismatch = PClassEquals(self.pclass(foo=u"bar")).match(self.pclass())
        self.expectThat(
            mismatch.describe(),
            Equals(
                u"field mismatch:\n"
                u"field: foo\n"
                u"reference = bar\n"
                u"actual    = <<missing>>\n"
            ),
        )

        # Expected value missing an attribute.
        mismatch = PClassEquals(self.pclass()).match(self.pclass(foo=u"baz"))
        self.expectThat(
            mismatch.describe(),
            Equals(
                u"field mismatch:\n"
                u"field: foo\n"
                u"reference = <<missing>>\n"
                u"actual    = baz\n"
            ),
        )

        # The matcher has a nice string representation.
        self.expectThat(
            str(PClassEquals(self.pclass(foo=u"bar"))),
            Equals("PClassEquals(pclass(foo=u'bar'))"),
        )



# Copyright Least Authority Enterprises.
# See LICENSE for details.

__all__ = [
    "strategies",
    "integration",
    "TestCase", "AsynchronousDeferredRunTest",
]

from ._testcase import TestCase, AsynchronousDeferredRunTest



# Copyright Least Authority Enterprises.
# See LICENSE for details.

"""
Tests for ``txkube.memory_kubernetes``.
"""

from zope.interface.verify import verifyClass

from hypothesis import given

from testtools.matchers import Equals, Is, IsInstance, AfterPreprocessing

from ..testing.integration import kubernetes_client_tests
from ..testing.strategies import iobjects
from ..testing import TestCase

from .. import memory_kubernetes
from .._memory import (
    _KubernetesState, IAgency, NullAgency, _incrementResourceVersion,
)


def get_kubernetes(case):
    """
    Create an in-memory test double provider of ``IKubernetes``.
    """
    return memory_kubernetes()



class KubernetesClientIntegrationTests(kubernetes_client_tests(get_kubernetes)):
    """
    Integration tests which interact with an in-memory-only Kubernetes
    deployment via ``txkube.memory_kubernetes``.
    """



class NullAgencyTests(TestCase):
    """
    Tests for ``NullAgency``.
    """
    def test_interface(self):
        """
        ``NullAgency`` implements ``IAgency``.
        """
        verifyClass(IAgency, NullAgency)


    @given(iobjects())
    def test_before_create(self, obj):
        """
        ``NullAgency.before_create`` returns the ``IObject`` passed to it with no
        modifications.
        """
        state = _KubernetesState()
        actual = NullAgency().before_create(state, obj)
        self.assertThat(actual, Equals(obj))


    @given(iobjects())
    def test_after_create(self, obj):
        """
        ``NullAgency.after_create`` returns the ``IObject`` passed to it with no
        modifications.
        """
        state = _KubernetesState()
        actual = NullAgency().after_create(state, obj)
        self.assertThat(actual, Equals(obj))


    @given(iobjects(), iobjects())
    def test_before_replace(self, old, new):
        """
        ``NullAgency.before_replace`` returns ``None``.
        """
        state = _KubernetesState()
        actual = NullAgency().before_replace(state, old, new)
        self.assertThat(actual, Is(None))



class IncrementResourceVersionTests(TestCase):
    """
    Tests for ``_incrementResourceVersion``.
    """
    def test_missing(self):
        """
        The next version after ``None`` is ``u"1"``.
        """
        version = _incrementResourceVersion(None)
        self.expectThat(version, IsInstance(unicode))
        self.expectThat(version, Equals(u"1"))


    def test_incremented(self):
        """
        Any version other than ``None`` is interpreted as a string representation
        of an integer and incremented.
        """
        version = _incrementResourceVersion(_incrementResourceVersion(None))
        updated = _incrementResourceVersion(version)
        self.expectThat(updated, IsInstance(unicode))
        self.expectThat(updated, AfterPreprocessing(int, Equals(int(version) + 1)))



# Copyright Least Authority Enterprises.
# See LICENSE for details.

"""
An in-memory implementation of the Kubernetes client interface.
"""

from json import dumps, loads

import attr

from pyrsistent import InvariantException, PClass, field, pset

from zope.interface import Interface, implementer

from twisted.python.url import URL

from twisted.web.http import CREATED, NOT_FOUND, OK

from eliot import start_action

from klein import Klein

from werkzeug.exceptions import NotFound

from treq.testing import RequestTraversalAgent

from . import (
    IKubernetes, KubernetesError, network_kubernetes,
    v1, v1beta1,
    iobject_from_raw, iobject_to_raw,
)


def memory_kubernetes():
    """
    Create an in-memory Kubernetes-alike service.

    This serves as a places to hold state for stateful Kubernetes interactions
    allowed by ``IKubernetesClient``.  Only clients created against the same
    instance will all share state.

    :return IKubernetes: The new Kubernetes-alike service.
    """
    return _MemoryKubernetes()


@implementer(IKubernetes)
class _MemoryKubernetes(object):
    """
    ``_MemoryKubernetes`` maintains state in-memory which approximates
    the state of a real Kubernetes deployment sufficiently to expose a
    subset of the external Kubernetes API.
    """
    def __init__(self):
        base_url = URL.fromText(u"https://kubernetes.example.invalid./")
        self._state = _KubernetesState()
        self._resource = _kubernetes_resource(self)
        self._kubernetes = network_kubernetes(
            base_url=base_url,
            agent=RequestTraversalAgent(self._resource),
        )


    def _state_changed(self, state):
        """
        The Kubernetes state has been changed.  Record the new version.

        The state is immutable so any changes must be represented as a brand
        new object.

        :param _KubernetesState state: The new state.
        """
        self._state = state


    def client(self, *args, **kwargs):
        """
        :return IKubernetesClient: A new client which interacts with this
            object rather than a real Kubernetes deployment.
        """
        return self._kubernetes.client(*args, **kwargs)



def _kubernetes_resource(memory_service):
    return _Kubernetes(memory_service).app.resource()



def _incrementResourceVersion(version):
    """
    Pyrsistent transformation function which can increment a
    ``v1.ObjectMeta.resourceVersion`` value (even if it was missing).

    :param version: The old version as a ``unicode`` string or ``None`` if
        there wasn't one.

    :return unicode: The new version, guaranteed to be greater than the old
        one.
    """
    if version is None:
        version = 0
    return u"{}".format(int(version) + 1)



def _transform_object(obj, *transformation):
    """
    Apply a pyrsistent transformation to an ``IObject``.

    In addition to the given transformation, the object's resourceVersion will
    be updated.

    :param IObject: obj: The object to transform.
    :param *transformation: Arguments like those to ``PClass.transform``.

    :return: The transformed object.
    """
    return obj.transform(
        [u"metadata", u"resourceVersion"],
        _incrementResourceVersion,
        *transformation
    )



# Help figure out what API group some types are associated with.
_groups = {
    v1beta1.Deployment: u"extensions",
    v1beta1.DeploymentList: u"extensions",
}



class IAgency(Interface):
    """
    An ``IAgency`` implementation can impress certain additional behaviors
    upon a ``_KubernetesState``.  The latter shall use methods of the former
    during state changes to give the former an opportunity to influence the
    outcome of the state change.
    """
    def before_create(state, obj):
        """
        This is called before an object is created.

        :param _KubernetesState state: The state in which the object is being
            created.

        :param IObject obj: A description of the object to be created.

        :return IObject: The object to really create.  Typically this is some
            transformation of ``obj`` (for example, with default values
            populated).
        """


    def after_create(state, obj):
        """
        This is called after an object has been created.

        :param _KubernetesState state: The state in which the object is being
            created.

        :param IObject obj: A description of the object created.  Regardless
            of the implementation of this method, this is the description
            which will be returned in the response to the create operation.

        :return IObject: The object to store in the state.  Typically this is
            some transformation of ``obj`` (for example, with an observed
            status attached)l.
        """


    def before_replace(state, old, new):
        """
        This is called before an existing object is replaced by a new one.

        :param _KubernetesState state: The state in which the object is being
            replaced.

        :param IObject old: A description of the object being replaced.

        :param IObject new: A description of the object to replace ``old``.

        :raise: Some exception to prevent the replacement from taking place.

        :return: ``None``
        """



@implementer(IAgency)
class NullAgency(object):
    """
    ``NullAgency`` does nothing.
    """
    def before_create(self, state, obj):
        return obj


    def after_create(self, state, obj):
        return obj


    def before_replace(self, state, old, new):
        pass



@implementer(IAgency)
class AdHocAgency(object):
    """
    ``AdHocAgency`` implements some object changes which I observed to happen
    on a real Kubernetes server while I was working on various parts of
    txkube.  No attempt at completeness attempted.  The system for selecting
    changes to implement is to run into important inconsistencies between this
    and a real Kubernetes while developing other features and then fix those
    inconsistencies.

    Perhaps in the future this will be replaced by something with less of an
    ad hoc nature.
    """
    def before_create(self, state, obj):
        return obj.fill_defaults()


    def after_create(self, state, obj):
        if isinstance(obj, v1beta1.Deployment):
            obj = _transform_object(
                obj,
                [u"metadata", u"annotations", u"deployment.kubernetes.io/revision"],
                u"1",
                [u"status"],
                {},
                [u"status", u"observedGeneration"],
                1,
                [u"status", u"unavailableReplicas"],
                1,
            )
        return obj


    def before_replace(self, state, old, new):
        if old.metadata.resourceVersion != new.metadata.resourceVersion:
            group = _groups.get(type(old), None)
            details = {
                u"group": group,
                u"kind": old.kind,
                u"name": old.metadata.name,
            }
            raise KubernetesError.object_modified(details)


class _KubernetesState(PClass):
    """
    ``_KubernetesState`` contains the canonical representation of internal
    state required to emulate a Kubernetes server.

    :ivar IAgency agency: Any behavior to apply to transformations of this
        state.
    """
    agency = field(initial=AdHocAgency())

    namespaces = field(initial=v1.NamespaceList())
    configmaps = field(initial=v1.ConfigMapList())
    services = field(initial=v1.ServiceList())
    pods = field(initial=v1.PodList())

    deployments = field(initial=v1beta1.DeploymentList())
    replicasets = field(initial=v1beta1.ReplicaSetList())


    def create(self, collection_name, obj):
        """
        Create a new object in the named collection.

        :param unicode collection_name: The name of the collection in which to
            create the object.

        :param IObject obj: A description of the object to create.

        :return _KubernetesState: A new state based on the current state but
            also containing ``obj``.
        """
        obj = self.agency.before_create(self, obj)
        new = self.agency.after_create(self, obj)
        updated = self.transform(
            [collection_name],
            lambda c: c.add(new),
        )
        return updated


    def replace(self, collection_name, old, new):
        """
        Replace an existing object with a new version of it.

        :param unicode collection_name: The name of the collection in which to
            replace an object.

        :param IObject old: A description of the object being replaced.

        :param IObject new: A description of the object to take the place of
            ``old``.

        :return _KubernetesState: A new state based on the current state but
            also containing ``obj``.
        """
        self.agency.before_replace(self, old, new)
        updated = self.transform(
            [collection_name],
            lambda c: c.replace(old, new),
        )
        return updated


    def delete(self, collection_name, obj):
        """
        Delete an existing object.

        :param unicode collection_name: The name of the collection from which
            to delete the object.

        :param IObject obj: A description of the object to delete.

        :return _KubernetesState: A new state based on the current state but
            not containing ``obj``.
        """
        updated = self.transform(
            [collection_name],
            lambda c: obj.delete_from(c),
        )
        return updated



def response(request, status, obj):
    """
    Generate a response.

    :param IRequest request: The request being responsed to.
    :param int status: The response status code to set.
    :param obj: Something JSON-dumpable to write into the response body.

    :return bytes: The response body to write out.  eg, return this from a
        *render_* method.
    """
    request.setResponseCode(status)
    request.responseHeaders.setRawHeaders(
        u"content-type", [u"application/json"],
    )
    return dumps(obj)



@attr.s(frozen=True)
class _Kubernetes(object):
    """
    A place to stick a bunch of Klein definitions.

    :ivar _MemoryKubernetes service: The Kubernetes-alike holding the
        in-memory Kubernetes state with which the API will be interacting.
    """
    service = attr.ib()

    # This could be a property except
    # https://github.com/hynek/attrs/issues/144
    def _get_state(self):
        return self.service._state

    def _set_state(self, value):
        self.service._state_changed(value)

    def _reduce_to_namespace(self, collection, namespace):
        # Unfortunately pset does not support transform. :( Use this more
        # verbose .set() operation.
        return collection.set(
            u"items",
            pset(obj for obj in collection.items if obj.metadata.namespace == namespace),
        )

    def _collection_by_name(self, collection_name):
        return getattr(self._get_state(), collection_name)

    def _list(self, request, namespace, collection_name):
        with start_action(action_type=u"memory:list", kind=collection_name):
            collection = self._collection_by_name(collection_name)
            if namespace is not None:
                collection = self._reduce_to_namespace(collection, namespace)
            return response(request, OK, iobject_to_raw(collection))

    def _get(self, request, collection_name, namespace, name):
        collection = self._collection_by_name(collection_name)
        if namespace is not None:
            collection = self._reduce_to_namespace(collection, namespace)
        try:
            obj = collection.item_by_name(name)
        except KeyError:
            raise KubernetesError.not_found({
                u"name": name,
                u"kind": collection_name,
                u"group": _groups.get(type(collection), None),
            })
        return response(request, OK, iobject_to_raw(obj))

    def _create(self, request, collection_name):
        with start_action(action_type=u"memory:create", kind=collection_name):
            obj = iobject_from_raw(loads(request.content.read()))
            try:
                state = self._get_state().create(collection_name, obj)
            except InvariantException:
                collection = getattr(self._get_state(), collection_name)
                details = {
                    u"name": obj.metadata.name,
                    u"kind": collection_name,
                    u"group": _groups.get(type(collection), None),
                }
                raise KubernetesError.already_exists(details)
            self._set_state(state)
            return response(request, CREATED, iobject_to_raw(obj))

    def _replace(self, request, collection_name, namespace, name):
        collection = self._collection_by_name(collection_name)
        if namespace is not None:
            collection = self._reduce_to_namespace(collection, namespace)
        old = collection.item_by_name(name)
        new = iobject_from_raw(loads(request.content.read()))
        try:
            state = self._get_state().replace(collection_name, old, new)
        except KubernetesError as e:
            return response(request, e.code, iobject_to_raw(e.status))

        self._set_state(state)
        return response(request, OK, iobject_to_raw(new))

    def _delete(self, request, collection_name, namespace, name):
        collection = self._collection_by_name(collection_name)
        if namespace is not None:
            collection = self._reduce_to_namespace(collection, namespace)
        try:
            obj = collection.item_by_name(name)
        except KeyError:
            raise KubernetesError.not_found({
                u"group": _groups.get(type(collection), None),
                u"kind": collection_name,
                u"name": name,
            })
        self._set_state(self._get_state().delete(collection_name, obj))
        return response(request, OK, iobject_to_raw(obj))

    app = Klein()
    @app.handle_errors(NotFound)
    def not_found(self, request, name):
        return response(
            request,
            NOT_FOUND,
            iobject_to_raw(v1.Status(
                status=u"Failure",
                message=u"the server could not find the requested resource",
                reason=u"NotFound",
                details={},
                metadata={},
                code=NOT_FOUND,
            )),
        )

    @app.handle_errors(KubernetesError)
    def object_not_found(self, request, reason):
        exc = reason.value
        return response(request, exc.code, iobject_to_raw(exc.status))

    with app.subroute(u"/api/v1") as app:
        @app.route(u"/namespaces", methods=[u"GET"])
        def list_namespaces(self, request):
            """
            Get all existing Namespaces.
            """
            return self._list(request, None, u"namespaces")

        @app.route(u"/namespaces/<namespace>", methods=[u"GET"])
        def get_namespace(self, request, namespace):
            """
            Get one Namespace by name.
            """
            return self._get(request, u"namespaces", None, namespace)

        @app.route(u"/namespaces/<namespace>", methods=[u"DELETE"])
        def delete_namespace(self, request, namespace):
            """
            Delete one Namespace by name.
            """
            return self._delete(
                request, u"namespaces", None, namespace,
            )

        @app.route(u"/namespaces", methods=[u"POST"])
        def create_namespace(self, request):
            """
            Create a new Namespace.
            """
            return self._create(request, u"namespaces")

        @app.route(u"/namespaces/<namespace>", methods=[u"PUT"])
        def replace_namespace(self, request, namespace):
            """
            Replace an existing Namespace.
            """
            return self._replace(
                request,
                u"namespaces",
                None,
                namespace,
            )

        @app.route(u"/namespaces/<namespace>/pods", methods=[u"POST"])
        def create_pod(self, request, namespace):
            """
            Create a new Pod.
            """
            return self._create(request, u"pods")

        @app.route(u"/namespaces/<namespace>/pods/<pod>", methods=[u"DELETE"])
        def delete_pod(self, request, namespace, pod):
            """
            Delete one Pod by name
            """
            return self._delete(request, u"pods", namespace, pod)

        @app.route(u"/pods", methods=[u"GET"])
        def list_pods(self, request):
            """
            Get all existing Pods.
            """
            return self._list(request, None, u"pods")

        @app.route(u"/namespaces/<namespace>/pods/<pod>", methods=[u"PUT"])
        def replace_pod(self, request, namespace, pod):
            """
            Replace an existing Pod.
            """
            return self._replace(request, u"pods", namespace, pod)

        @app.route(u"/namespaces/<namespace>/pods/<pod>", methods=[u"GET"])
        def get_pod(self, request, namespace, pod):
            """
            Get one Pod by name.
            """
            return self._get(request, u"pods", namespace, pod)

        @app.route(u"/configmaps", methods=[u"GET"])
        def list_configmaps(self, request):
            """
            Get all existing ConfigMaps.
            """
            return self._list(request, None, u"configmaps")

        @app.route(u"/namespaces/<namespace>/configmaps/<configmap>", methods=[u"GET"])
        def get_configmap(self, request, namespace, configmap):
            """
            Get one ConfigMap by name.
            """
            return self._get(request, u"configmaps", namespace, configmap)

        @app.route(u"/namespaces/<namespace>/configmaps", methods=[u"POST"])
        def create_configmap(self, request, namespace):
            """
            Create a new ConfigMap.
            """
            return self._create(request, u"configmaps")

        @app.route(u"/namespaces/<namespace>/configmaps/<configmap>", methods=[u"PUT"])
        def replace_configmap(self, request, namespace, configmap):
            """
            Replace an existing ConfigMap.
            """
            return self._replace(
                request,
                u"configmaps",
                namespace,
                configmap,
            )

        @app.route(u"/namespaces/<namespace>/configmaps/<configmap>", methods=[u"DELETE"])
        def delete_configmap(self, request, namespace, configmap):
            """
            Delete one ConfigMap by name.
            """
            return self._delete(
                request, u"configmaps", namespace, configmap,
            )

        @app.route(u"/namespaces/<namespace>/services", methods=[u"POST"])
        def create_service(self, request, namespace):
            """
            Create a new Service.
            """
            return self._create(request, u"services")

        @app.route(u"/namespaces/<namespace>/services/<service>", methods=[u"PUT"])
        def replace_service(self, request, namespace, service):
            """
            Replace an existing Service.
            """
            return self._replace(
                request,
                u"services",
                namespace,
                service,
            )

        @app.route(u"/services", methods=[u"GET"])
        def list_services(self, request):
            """
            Get all existing Services.
            """
            return self._list(request, None, u"services")

        @app.route(u"/namespaces/<namespace>/services/<service>", methods=[u"GET"])
        def get_service(self, request, namespace, service):
            """
            Get one Service by name.
            """
            return self._get(
                request,
                u"services",
                namespace,
                service,
            )

        @app.route(u"/namespaces/<namespace>/services/<service>", methods=[u"DELETE"])
        def delete_service(self, request, namespace, service):
            """
            Delete one Service by name.
            """
            return self._delete(
                request, u"services", namespace, service,
            )

    with app.subroute(u"/apis/extensions/v1beta1") as app:
        @app.route(u"/namespaces/<namespace>/deployments", methods=[u"POST"])
        def create_deployment(self, request, namespace):
            """
            Create a new Deployment.
            """
            return self._create(
                request,
                u"deployments",
            )

        @app.route(u"/namespaces/<namespace>/deployments/<deployment>", methods=[u"PUT"])
        def replace_deployment(self, request, namespace, deployment):
            """
            Replace an existing Deployment.
            """
            return self._replace(
                request,
                u"deployments",
                namespace,
                deployment,
            )

        @app.route(u"/deployments", methods=[u"GET"])
        def list_deployments(self, request):
            """
            Get all existing Deployments.
            """
            return self._list(request, None, u"deployments")

        @app.route(u"/namespaces/<namespace>/deployments/<deployment>", methods=[u"GET"])
        def get_deployment(self, request, namespace, deployment):
            """
            Get one Deployment by name.
            """
            return self._get(
                request,
                u"deployments",
                namespace,
                deployment,
            )

        @app.route(u"/namespaces/<namespace>/deployments/<deployment>", methods=[u"DELETE"])
        def delete_deployment(self, request, namespace, deployment):
            """
            Delete one Deployment by name.
            """
            return self._delete(
                request, u"deployments", namespace, deployment,
            )

        @app.route(u"/namespaces/<namespace>/replicasets", methods=[u"POST"])
        def create_replicaset(self, request, namespace):
            """
            Create a new ReplicaSet.
            """
            return self._create(
                request,
                u"replicasets",
            )

        @app.route(u"/namespaces/<namespace>/replicasets/<replicaset>", methods=[u"PUT"])
        def replace_replicaset(self, request, namespace, replicaset):
            """
            Replace an existing ReplicaSet.
            """
            return self._replace(
                request,
                u"replicasets",
                namespace,
                replicaset,
            )

        @app.route(u"/replicasets", methods=[u"GET"])
        def list_replicasets(self, request):
            """
            Get all existing ReplicaSets.
            """
            return self._list(request, None, u"replicasets")

        @app.route(u"/namespaces/<namespace>/replicasets/<replicaset>", methods=[u"GET"])
        def get_replicaset(self, request, namespace, replicaset):
            """
            Get one ReplicaSet by name.
            """
            return self._get(
                request,
                u"replicasets",
                namespace,
                replicaset,
            )

        @app.route(u"/namespaces/<namespace>/replicasets/<replicaset>", methods=[u"DELETE"])
        def delete_replicaset(self, request, namespace, replicaset):
            """
            Delete one ReplicaSet by name.
            """
            return self._delete(
                request, u"replicasets", namespace, replicaset,
            )
